from __future__ import unicode_literals

import datetime
import time
import boto.kinesis
import re
import six
import itertools

from operator import attrgetter
from hashlib import md5

from moto.compat import OrderedDict
from moto.core import BaseBackend, BaseModel
from .exceptions import StreamNotFoundError, ShardNotFoundError, ResourceInUseError, \
    ResourceNotFoundError, InvalidArgumentError
from .utils import compose_shard_iterator, compose_new_shard_iterator, decompose_shard_iterator


class Record(BaseModel):

    def __init__(self, partition_key, data, sequence_number, explicit_hash_key):
        self.partition_key = partition_key
        self.data = data
        self.sequence_number = sequence_number
        self.explicit_hash_key = explicit_hash_key

    def to_json(self):
        return {
            "Data": self.data,
            "PartitionKey": self.partition_key,
            "SequenceNumber": str(self.sequence_number),
        }


class Shard(BaseModel):

    def __init__(self, shard_id, starting_hash, ending_hash):
        self._shard_id = shard_id
        self.starting_hash = starting_hash
        self.ending_hash = ending_hash
        self.records = OrderedDict()

    @property
    def shard_id(self):
        return "shardId-{0}".format(str(self._shard_id).zfill(12))

    def get_records(self, last_sequence_id, limit):
        last_sequence_id = int(last_sequence_id)
        results = []

        for sequence_number, record in self.records.items():
            if sequence_number > last_sequence_id:
                results.append(record)
                last_sequence_id = sequence_number

            if len(results) == limit:
                break

        return results, last_sequence_id

    def put_record(self, partition_key, data, explicit_hash_key):
        # Note: this function is not safe for concurrency
        if self.records:
            last_sequence_number = self.get_max_sequence_number()
        else:
            last_sequence_number = 0
        sequence_number = last_sequence_number + 1
        self.records[sequence_number] = Record(
            partition_key, data, sequence_number, explicit_hash_key)
        return sequence_number

    def get_min_sequence_number(self):
        if self.records:
            return list(self.records.keys())[0]
        return 0

    def get_max_sequence_number(self):
        if self.records:
            return list(self.records.keys())[-1]
        return 0

    def to_json(self):
        return {
            "HashKeyRange": {
                "EndingHashKey": str(self.ending_hash),
                "StartingHashKey": str(self.starting_hash)
            },
            "SequenceNumberRange": {
                "EndingSequenceNumber": self.get_max_sequence_number(),
                "StartingSequenceNumber": self.get_min_sequence_number(),
            },
            "ShardId": self.shard_id
        }


class Stream(BaseModel):

    def __init__(self, stream_name, shard_count, region):
        self.stream_name = stream_name
        self.shard_count = shard_count
        self.region = region
        self.account_number = "123456789012"
        self.shards = {}
        self.tags = {}

        if six.PY3:
            izip_longest = itertools.zip_longest
        else:
            izip_longest = itertools.izip_longest

        for index, start, end in izip_longest(range(shard_count),
                                              range(0, 2**128, 2 **
                                                    128 // shard_count),
                                              range(2**128 // shard_count, 2 **
                                                    128, 2**128 // shard_count),
                                              fillvalue=2**128):
            shard = Shard(index, start, end)
            self.shards[shard.shard_id] = shard

    @property
    def arn(self):
        return "arn:aws:kinesis:{region}:{account_number}:{stream_name}".format(
            region=self.region,
            account_number=self.account_number,
            stream_name=self.stream_name
        )

    def get_shard(self, shard_id):
        if shard_id in self.shards:
            return self.shards[shard_id]
        else:
            raise ShardNotFoundError(shard_id)

    def get_shard_for_key(self, partition_key, explicit_hash_key):
        if not isinstance(partition_key, six.string_types):
            raise InvalidArgumentError("partition_key")
        if len(partition_key) > 256:
            raise InvalidArgumentError("partition_key")

        if explicit_hash_key:
            if not isinstance(explicit_hash_key, six.string_types):
                raise InvalidArgumentError("explicit_hash_key")

            key = int(explicit_hash_key)

            if key >= 2**128:
                raise InvalidArgumentError("explicit_hash_key")

        else:
            key = int(md5(partition_key.encode('utf-8')).hexdigest(), 16)

        for shard in self.shards.values():
            if shard.starting_hash <= key < shard.ending_hash:
                return shard

    def put_record(self, partition_key, explicit_hash_key, sequence_number_for_ordering, data):
        shard = self.get_shard_for_key(partition_key, explicit_hash_key)

        sequence_number = shard.put_record(
            partition_key, data, explicit_hash_key)
        return sequence_number, shard.shard_id

    def to_json(self):
        return {
            "StreamDescription": {
                "StreamARN": self.arn,
                "StreamName": self.stream_name,
                "StreamStatus": "ACTIVE",
                "HasMoreShards": False,
                "Shards": [shard.to_json() for shard in self.shards.values()],
            }
        }

    @classmethod
    def create_from_cloudformation_json(cls, resource_name, cloudformation_json, region_name):
        properties = cloudformation_json['Properties']
        region = properties.get('Region', 'us-east-1')
        shard_count = properties.get('ShardCount', 1)
        return Stream(properties['Name'], shard_count, region)


class FirehoseRecord(BaseModel):

    def __init__(self, record_data):
        self.record_id = 12345678
        self.record_data = record_data


class DeliveryStream(BaseModel):

    def __init__(self, stream_name, **stream_kwargs):
        self.name = stream_name
        self.redshift_username = stream_kwargs.get('redshift_username')
        self.redshift_password = stream_kwargs.get('redshift_password')
        self.redshift_jdbc_url = stream_kwargs.get('redshift_jdbc_url')
        self.redshift_role_arn = stream_kwargs.get('redshift_role_arn')
        self.redshift_copy_command = stream_kwargs.get('redshift_copy_command')

        self.s3_role_arn = stream_kwargs.get('s3_role_arn')
        self.s3_bucket_arn = stream_kwargs.get('s3_bucket_arn')
        self.s3_prefix = stream_kwargs.get('s3_prefix')
        self.s3_compression_format = stream_kwargs.get(
            's3_compression_format', 'UNCOMPRESSED')
        self.s3_buffering_hings = stream_kwargs.get('s3_buffering_hings')

        self.redshift_s3_role_arn = stream_kwargs.get('redshift_s3_role_arn')
        self.redshift_s3_bucket_arn = stream_kwargs.get(
            'redshift_s3_bucket_arn')
        self.redshift_s3_prefix = stream_kwargs.get('redshift_s3_prefix')
        self.redshift_s3_compression_format = stream_kwargs.get(
            'redshift_s3_compression_format', 'UNCOMPRESSED')
        self.redshift_s3_buffering_hings = stream_kwargs.get(
            'redshift_s3_buffering_hings')

        self.records = []
        self.status = 'ACTIVE'
        self.create_at = datetime.datetime.utcnow()
        self.last_updated = datetime.datetime.utcnow()

    @property
    def arn(self):
        return 'arn:aws:firehose:us-east-1:123456789012:deliverystream/{0}'.format(self.name)

    def destinations_to_dict(self):
        if self.s3_role_arn:
            return [{
                'DestinationId': 'string',
                'S3DestinationDescription': {
                    'RoleARN': self.s3_role_arn,
                    'BucketARN': self.s3_bucket_arn,
                    'Prefix': self.s3_prefix,
                    'BufferingHints': self.s3_buffering_hings,
                    'CompressionFormat': self.s3_compression_format,
                }
            }]
        else:
            return [{
                    "DestinationId": "string",
                    "RedshiftDestinationDescription": {
                        "ClusterJDBCURL": self.redshift_jdbc_url,
                        "CopyCommand": self.redshift_copy_command,
                        "RoleARN": self.redshift_role_arn,
                        "S3DestinationDescription": {
                            "BucketARN": self.redshift_s3_bucket_arn,
                            "BufferingHints": self.redshift_s3_buffering_hings,
                            "CompressionFormat": self.redshift_s3_compression_format,
                            "Prefix": self.redshift_s3_prefix,
                            "RoleARN": self.redshift_s3_role_arn
                        },
                        "Username": self.redshift_username,
                    },
                    }
                    ]

    def to_dict(self):
        return {
            "DeliveryStreamDescription": {
                "CreateTimestamp": time.mktime(self.create_at.timetuple()),
                "DeliveryStreamARN": self.arn,
                "DeliveryStreamName": self.name,
                "DeliveryStreamStatus": self.status,
                "Destinations": self.destinations_to_dict(),
                "HasMoreDestinations": False,
                "LastUpdateTimestamp": time.mktime(self.last_updated.timetuple()),
                "VersionId": "string",
            }
        }

    def put_record(self, record_data):
        record = FirehoseRecord(record_data)
        self.records.append(record)
        return record


class KinesisBackend(BaseBackend):

    def __init__(self):
        self.streams = OrderedDict()
        self.delivery_streams = {}

    def create_stream(self, stream_name, shard_count, region):
        if stream_name in self.streams:
            raise ResourceInUseError(stream_name)
        stream = Stream(stream_name, shard_count, region)
        self.streams[stream_name] = stream
        return stream

    def describe_stream(self, stream_name):
        if stream_name in self.streams:
            return self.streams[stream_name]
        else:
            raise StreamNotFoundError(stream_name)

    def list_streams(self):
        return self.streams.values()

    def delete_stream(self, stream_name):
        if stream_name in self.streams:
            return self.streams.pop(stream_name)
        raise StreamNotFoundError(stream_name)

    def get_shard_iterator(self, stream_name, shard_id, shard_iterator_type, starting_sequence_number):
        # Validate params
        stream = self.describe_stream(stream_name)
        shard = stream.get_shard(shard_id)

        shard_iterator = compose_new_shard_iterator(
            stream_name, shard, shard_iterator_type, starting_sequence_number
        )
        return shard_iterator

    def get_records(self, shard_iterator, limit):
        decomposed = decompose_shard_iterator(shard_iterator)
        stream_name, shard_id, last_sequence_id = decomposed

        stream = self.describe_stream(stream_name)
        shard = stream.get_shard(shard_id)

        records, last_sequence_id = shard.get_records(last_sequence_id, limit)

        next_shard_iterator = compose_shard_iterator(
            stream_name, shard, last_sequence_id)

        return next_shard_iterator, records

    def put_record(self, stream_name, partition_key, explicit_hash_key, sequence_number_for_ordering, data):
        stream = self.describe_stream(stream_name)

        sequence_number, shard_id = stream.put_record(
            partition_key, explicit_hash_key, sequence_number_for_ordering, data
        )

        return sequence_number, shard_id

    def put_records(self, stream_name, records):
        stream = self.describe_stream(stream_name)

        response = {
            "FailedRecordCount": 0,
            "Records": []
        }

        for record in records:
            partition_key = record.get("PartitionKey")
            explicit_hash_key = record.get("ExplicitHashKey")
            data = record.get("Data")

            sequence_number, shard_id = stream.put_record(
                partition_key, explicit_hash_key, None, data
            )
            response['Records'].append({
                "SequenceNumber": sequence_number,
                "ShardId": shard_id
            })

        return response

    def split_shard(self, stream_name, shard_to_split, new_starting_hash_key):
        stream = self.describe_stream(stream_name)

        if shard_to_split not in stream.shards:
            raise ResourceNotFoundError(shard_to_split)

        if not re.match(r'0|([1-9]\d{0,38})', new_starting_hash_key):
            raise InvalidArgumentError(new_starting_hash_key)
        new_starting_hash_key = int(new_starting_hash_key)

        shard = stream.shards[shard_to_split]

        last_id = sorted(stream.shards.values(),
                         key=attrgetter('_shard_id'))[-1]._shard_id

        if shard.starting_hash < new_starting_hash_key < shard.ending_hash:
            new_shard = Shard(
                last_id + 1, new_starting_hash_key, shard.ending_hash)
            shard.ending_hash = new_starting_hash_key
            stream.shards[new_shard.shard_id] = new_shard
        else:
            raise InvalidArgumentError(new_starting_hash_key)

        records = shard.records
        shard.records = OrderedDict()

        for index in records:
            record = records[index]
            stream.put_record(
                record.partition_key, record.explicit_hash_key, None, record.data
            )

    def merge_shards(self, stream_name, shard_to_merge, adjacent_shard_to_merge):
        stream = self.describe_stream(stream_name)

        if shard_to_merge not in stream.shards:
            raise ResourceNotFoundError(shard_to_merge)

        if adjacent_shard_to_merge not in stream.shards:
            raise ResourceNotFoundError(adjacent_shard_to_merge)

        shard1 = stream.shards[shard_to_merge]
        shard2 = stream.shards[adjacent_shard_to_merge]

        if shard1.ending_hash == shard2.starting_hash:
            shard1.ending_hash = shard2.ending_hash
        elif shard2.ending_hash == shard1.starting_hash:
            shard1.starting_hash = shard2.starting_hash
        else:
            raise InvalidArgumentError(adjacent_shard_to_merge)

        del stream.shards[shard2.shard_id]
        for index in shard2.records:
            record = shard2.records[index]
            shard1.put_record(record.partition_key,
                              record.data, record.explicit_hash_key)

    ''' Firehose '''

    def create_delivery_stream(self, stream_name, **stream_kwargs):
        stream = DeliveryStream(stream_name, **stream_kwargs)
        self.delivery_streams[stream_name] = stream
        return stream

    def get_delivery_stream(self, stream_name):
        if stream_name in self.delivery_streams:
            return self.delivery_streams[stream_name]
        else:
            raise StreamNotFoundError(stream_name)

    def list_delivery_streams(self):
        return self.delivery_streams.values()

    def delete_delivery_stream(self, stream_name):
        self.delivery_streams.pop(stream_name)

    def put_firehose_record(self, stream_name, record_data):
        stream = self.get_delivery_stream(stream_name)
        record = stream.put_record(record_data)
        return record

    def list_tags_for_stream(self, stream_name, exclusive_start_tag_key=None, limit=None):
        stream = self.describe_stream(stream_name)

        tags = []
        result = {
            'HasMoreTags': False,
            'Tags': tags
        }
        for key, val in sorted(stream.tags.items(), key=lambda x: x[0]):
            if limit and len(tags) >= limit:
                result['HasMoreTags'] = True
                break
            if exclusive_start_tag_key and key < exclusive_start_tag_key:
                continue

            tags.append({
                'Key': key,
                'Value': val
            })

        return result

    def add_tags_to_stream(self, stream_name, tags):
        stream = self.describe_stream(stream_name)
        stream.tags.update(tags)

    def remove_tags_from_stream(self, stream_name, tag_keys):
        stream = self.describe_stream(stream_name)
        for key in tag_keys:
            if key in stream.tags:
                del stream.tags[key]


kinesis_backends = {}
for region in boto.kinesis.regions():
    kinesis_backends[region.name] = KinesisBackend()



from __future__ import unicode_literals

import json
import sure  # noqa

import moto.server as server
from moto import mock_kinesis

'''
Test the different server responses
'''


@mock_kinesis
def test_list_streams():
    backend = server.create_backend_app("kinesis")
    test_client = backend.test_client()

    res = test_client.get('/?Action=ListStreams')

    json_data = json.loads(res.data.decode("utf-8"))
    json_data.should.equal({
        "HasMoreStreams": False,
        "StreamNames": [],
    })



from __future__ import unicode_literals
# Ensure 'assert_raises' context manager support for Python 2.6
import tests.backport_assert_raises
from nose.tools import assert_raises

import boto
import six
import sure  # noqa

from boto.exception import EC2ResponseError
from moto import mock_ec2_deprecated


@mock_ec2_deprecated
def test_key_pairs_empty():
    conn = boto.connect_ec2('the_key', 'the_secret')
    assert len(conn.get_all_key_pairs()) == 0


@mock_ec2_deprecated
def test_key_pairs_invalid_id():
    conn = boto.connect_ec2('the_key', 'the_secret')

    with assert_raises(EC2ResponseError) as cm:
        conn.get_all_key_pairs('foo')
    cm.exception.code.should.equal('InvalidKeyPair.NotFound')
    cm.exception.status.should.equal(400)
    cm.exception.request_id.should_not.be.none


@mock_ec2_deprecated
def test_key_pairs_create():
    conn = boto.connect_ec2('the_key', 'the_secret')

    with assert_raises(EC2ResponseError) as ex:
        kp = conn.create_key_pair('foo', dry_run=True)
    ex.exception.error_code.should.equal('DryRunOperation')
    ex.exception.status.should.equal(400)
    ex.exception.message.should.equal(
        'An error occurred (DryRunOperation) when calling the CreateKeyPair operation: Request would have succeeded, but DryRun flag is set')

    kp = conn.create_key_pair('foo')
    assert kp.material.startswith('---- BEGIN RSA PRIVATE KEY ----')
    kps = conn.get_all_key_pairs()
    assert len(kps) == 1
    assert kps[0].name == 'foo'


@mock_ec2_deprecated
def test_key_pairs_create_two():
    conn = boto.connect_ec2('the_key', 'the_secret')
    kp = conn.create_key_pair('foo')
    kp = conn.create_key_pair('bar')
    assert kp.material.startswith('---- BEGIN RSA PRIVATE KEY ----')
    kps = conn.get_all_key_pairs()
    kps.should.have.length_of(2)
    [i.name for i in kps].should.contain('foo')
    [i.name for i in kps].should.contain('bar')
    kps = conn.get_all_key_pairs('foo')
    kps.should.have.length_of(1)
    kps[0].name.should.equal('foo')


@mock_ec2_deprecated
def test_key_pairs_create_exist():
    conn = boto.connect_ec2('the_key', 'the_secret')
    kp = conn.create_key_pair('foo')
    assert kp.material.startswith('---- BEGIN RSA PRIVATE KEY ----')
    assert len(conn.get_all_key_pairs()) == 1

    with assert_raises(EC2ResponseError) as cm:
        conn.create_key_pair('foo')
    cm.exception.code.should.equal('InvalidKeyPair.Duplicate')
    cm.exception.status.should.equal(400)
    cm.exception.request_id.should_not.be.none


@mock_ec2_deprecated
def test_key_pairs_delete_no_exist():
    conn = boto.connect_ec2('the_key', 'the_secret')
    assert len(conn.get_all_key_pairs()) == 0
    r = conn.delete_key_pair('foo')
    r.should.be.ok


@mock_ec2_deprecated
def test_key_pairs_delete_exist():
    conn = boto.connect_ec2('the_key', 'the_secret')
    conn.create_key_pair('foo')

    with assert_raises(EC2ResponseError) as ex:
        r = conn.delete_key_pair('foo', dry_run=True)
    ex.exception.error_code.should.equal('DryRunOperation')
    ex.exception.status.should.equal(400)
    ex.exception.message.should.equal(
        'An error occurred (DryRunOperation) when calling the DeleteKeyPair operation: Request would have succeeded, but DryRun flag is set')

    r = conn.delete_key_pair('foo')
    r.should.be.ok
    assert len(conn.get_all_key_pairs()) == 0


@mock_ec2_deprecated
def test_key_pairs_import():
    conn = boto.connect_ec2('the_key', 'the_secret')

    with assert_raises(EC2ResponseError) as ex:
        kp = conn.import_key_pair('foo', b'content', dry_run=True)
    ex.exception.error_code.should.equal('DryRunOperation')
    ex.exception.status.should.equal(400)
    ex.exception.message.should.equal(
        'An error occurred (DryRunOperation) when calling the ImportKeyPair operation: Request would have succeeded, but DryRun flag is set')

    kp = conn.import_key_pair('foo', b'content')
    assert kp.name == 'foo'
    kps = conn.get_all_key_pairs()
    assert len(kps) == 1
    assert kps[0].name == 'foo'


@mock_ec2_deprecated
def test_key_pairs_import_exist():
    conn = boto.connect_ec2('the_key', 'the_secret')
    kp = conn.import_key_pair('foo', b'content')
    assert kp.name == 'foo'
    assert len(conn.get_all_key_pairs()) == 1

    with assert_raises(EC2ResponseError) as cm:
        conn.create_key_pair('foo')
    cm.exception.code.should.equal('InvalidKeyPair.Duplicate')
    cm.exception.status.should.equal(400)
    cm.exception.request_id.should_not.be.none


@mock_ec2_deprecated
def test_key_pair_filters():
    conn = boto.connect_ec2('the_key', 'the_secret')

    _ = conn.create_key_pair('kpfltr1')
    kp2 = conn.create_key_pair('kpfltr2')
    kp3 = conn.create_key_pair('kpfltr3')

    kp_by_name = conn.get_all_key_pairs(
        filters={'key-name': 'kpfltr2'})
    set([kp.name for kp in kp_by_name]
        ).should.equal(set([kp2.name]))

    kp_by_name = conn.get_all_key_pairs(
        filters={'fingerprint': kp3.fingerprint})
    set([kp.name for kp in kp_by_name]
        ).should.equal(set([kp3.name]))



from __future__ import unicode_literals
import boto3
from freezegun import freeze_time
import sure  # noqa
import re

from moto import mock_opsworks


@freeze_time("2015-01-01")
@mock_opsworks
def test_create_layer_response():
    client = boto3.client('opsworks', region_name='us-east-1')
    stack_id = client.create_stack(
        Name="test_stack_1",
        Region="us-east-1",
        ServiceRoleArn="service_arn",
        DefaultInstanceProfileArn="profile_arn"
    )['StackId']

    response = client.create_layer(
        StackId=stack_id,
        Type="custom",
        Name="TestLayer",
        Shortname="TestLayerShortName"
    )

    response.should.contain("LayerId")

    second_stack_id = client.create_stack(
        Name="test_stack_2",
        Region="us-east-1",
        ServiceRoleArn="service_arn",
        DefaultInstanceProfileArn="profile_arn"
    )['StackId']

    response = client.create_layer(
        StackId=second_stack_id,
        Type="custom",
        Name="TestLayer",
        Shortname="TestLayerShortName"
    )

    response.should.contain("LayerId")

    # ClientError
    client.create_layer.when.called_with(
        StackId=stack_id,
        Type="custom",
        Name="TestLayer",
        Shortname="_"
    ).should.throw(
        Exception, re.compile(r'already a layer named "TestLayer"')
    )
    # ClientError
    client.create_layer.when.called_with(
        StackId=stack_id,
        Type="custom",
        Name="_",
        Shortname="TestLayerShortName"
    ).should.throw(
        Exception, re.compile(
            r'already a layer with shortname "TestLayerShortName"')
    )


@freeze_time("2015-01-01")
@mock_opsworks
def test_describe_layers():
    client = boto3.client('opsworks', region_name='us-east-1')
    stack_id = client.create_stack(
        Name="test_stack_1",
        Region="us-east-1",
        ServiceRoleArn="service_arn",
        DefaultInstanceProfileArn="profile_arn"
    )['StackId']
    layer_id = client.create_layer(
        StackId=stack_id,
        Type="custom",
        Name="TestLayer",
        Shortname="TestLayerShortName"
    )['LayerId']

    rv1 = client.describe_layers(StackId=stack_id)
    rv2 = client.describe_layers(LayerIds=[layer_id])
    rv1['Layers'].should.equal(rv2['Layers'])

    rv1['Layers'][0]['Name'].should.equal("TestLayer")



import random
import string

from six.moves.urllib.parse import urlparse


def region_from_glacier_url(url):
    domain = urlparse(url).netloc

    if '.' in domain:
        return domain.split(".")[1]
    else:
        return 'us-east-1'


def vault_from_glacier_url(full_url):
    return full_url.split("/")[-1]


def get_job_id():
    return ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(92))



from __future__ import unicode_literals

import json
import sure  # noqa

import moto.server as server
from moto import mock_kms

'''
Test the different server responses
'''


@mock_kms
def test_list_keys():
    backend = server.create_backend_app("kms")
    test_client = backend.test_client()

    res = test_client.get('/?Action=ListKeys')

    json.loads(res.data.decode("utf-8")).should.equal({
        "Keys": [],
        "NextMarker": None,
        "Truncated": False,
    })



from __future__ import unicode_literals

import json

import dicttoxml
from jinja2 import Template
from six import iteritems

from moto.core.responses import BaseResponse
from .models import redshift_backends


def convert_json_error_to_xml(json_error):
    error = json.loads(json_error)
    code = error['Error']['Code']
    message = error['Error']['Message']
    template = Template("""
        <RedshiftClientError>
            <Error>
              <Code>{{ code }}</Code>
              <Message>{{ message }}</Message>
              <Type>Sender</Type>
            </Error>
            <RequestId>6876f774-7273-11e4-85dc-39e55ca848d1</RequestId>
        </RedshiftClientError>""")
    return template.render(code=code, message=message)


class RedshiftResponse(BaseResponse):

    @property
    def redshift_backend(self):
        return redshift_backends[self.region]

    def get_response(self, response):
        if self.request_json:
            return json.dumps(response)
        else:
            xml = dicttoxml.dicttoxml(response, attr_type=False, root=False)
            return xml.decode("utf-8")

    def call_action(self):
        status, headers, body = super(RedshiftResponse, self).call_action()
        if status >= 400 and not self.request_json:
            body = convert_json_error_to_xml(body)
        return status, headers, body

    def unpack_complex_list_params(self, label, names):
        unpacked_list = list()
        count = 1
        while self._get_param('{0}.{1}.{2}'.format(label, count, names[0])):
            param = dict()
            for i in range(len(names)):
                param[names[i]] = self._get_param(
                    '{0}.{1}.{2}'.format(label, count, names[i]))
            unpacked_list.append(param)
            count += 1
        return unpacked_list

    def unpack_list_params(self, label):
        unpacked_list = list()
        count = 1
        while self._get_param('{0}.{1}'.format(label, count)):
            unpacked_list.append(self._get_param(
                '{0}.{1}'.format(label, count)))
            count += 1
        return unpacked_list

    def _get_cluster_security_groups(self):
        cluster_security_groups = self._get_multi_param('ClusterSecurityGroups.member')
        if not cluster_security_groups:
            cluster_security_groups = self._get_multi_param('ClusterSecurityGroups.ClusterSecurityGroupName')
        return cluster_security_groups

    def _get_vpc_security_group_ids(self):
        vpc_security_group_ids = self._get_multi_param('VpcSecurityGroupIds.member')
        if not vpc_security_group_ids:
            vpc_security_group_ids = self._get_multi_param('VpcSecurityGroupIds.VpcSecurityGroupId')
        return vpc_security_group_ids

    def _get_subnet_ids(self):
        subnet_ids = self._get_multi_param('SubnetIds.member')
        if not subnet_ids:
            subnet_ids = self._get_multi_param('SubnetIds.SubnetIdentifier')
        return subnet_ids

    def create_cluster(self):
        cluster_kwargs = {
            "cluster_identifier": self._get_param('ClusterIdentifier'),
            "node_type": self._get_param('NodeType'),
            "master_username": self._get_param('MasterUsername'),
            "master_user_password": self._get_param('MasterUserPassword'),
            "db_name": self._get_param('DBName'),
            "cluster_type": self._get_param('ClusterType'),
            "cluster_security_groups": self._get_cluster_security_groups(),
            "vpc_security_group_ids": self._get_vpc_security_group_ids(),
            "cluster_subnet_group_name": self._get_param('ClusterSubnetGroupName'),
            "availability_zone": self._get_param('AvailabilityZone'),
            "preferred_maintenance_window": self._get_param('PreferredMaintenanceWindow'),
            "cluster_parameter_group_name": self._get_param('ClusterParameterGroupName'),
            "automated_snapshot_retention_period": self._get_int_param('AutomatedSnapshotRetentionPeriod'),
            "port": self._get_int_param('Port'),
            "cluster_version": self._get_param('ClusterVersion'),
            "allow_version_upgrade": self._get_bool_param('AllowVersionUpgrade'),
            "number_of_nodes": self._get_int_param('NumberOfNodes'),
            "publicly_accessible": self._get_param("PubliclyAccessible"),
            "encrypted": self._get_param("Encrypted"),
            "region_name": self.region,
            "tags": self.unpack_complex_list_params('Tags.Tag', ('Key', 'Value'))
        }
        cluster = self.redshift_backend.create_cluster(**cluster_kwargs).to_json()
        cluster['ClusterStatus'] = 'creating'
        return self.get_response({
            "CreateClusterResponse": {
                "CreateClusterResult": {
                    "Cluster": cluster,
                },
                "ResponseMetadata": {
                    "RequestId": "384ac68d-3775-11df-8963-01868b7c937a",
                }
            }
        })

    def restore_from_cluster_snapshot(self):
        restore_kwargs = {
            "snapshot_identifier": self._get_param('SnapshotIdentifier'),
            "cluster_identifier": self._get_param('ClusterIdentifier'),
            "port": self._get_int_param('Port'),
            "availability_zone": self._get_param('AvailabilityZone'),
            "allow_version_upgrade": self._get_bool_param(
                'AllowVersionUpgrade'),
            "cluster_subnet_group_name": self._get_param(
                'ClusterSubnetGroupName'),
            "publicly_accessible": self._get_param("PubliclyAccessible"),
            "cluster_parameter_group_name": self._get_param(
                'ClusterParameterGroupName'),
            "cluster_security_groups": self._get_cluster_security_groups(),
            "vpc_security_group_ids": self._get_vpc_security_group_ids(),
            "preferred_maintenance_window": self._get_param(
                'PreferredMaintenanceWindow'),
            "automated_snapshot_retention_period": self._get_int_param(
                'AutomatedSnapshotRetentionPeriod'),
            "region_name": self.region,
        }
        cluster = self.redshift_backend.restore_from_cluster_snapshot(**restore_kwargs).to_json()
        cluster['ClusterStatus'] = 'creating'
        return self.get_response({
            "RestoreFromClusterSnapshotResponse": {
                "RestoreFromClusterSnapshotResult": {
                    "Cluster": cluster,
                },
                "ResponseMetadata": {
                    "RequestId": "384ac68d-3775-11df-8963-01868b7c937a",
                }
            }
        })

    def describe_clusters(self):
        cluster_identifier = self._get_param("ClusterIdentifier")
        clusters = self.redshift_backend.describe_clusters(cluster_identifier)

        return self.get_response({
            "DescribeClustersResponse": {
                "DescribeClustersResult": {
                    "Clusters": [cluster.to_json() for cluster in clusters]
                },
                "ResponseMetadata": {
                    "RequestId": "384ac68d-3775-11df-8963-01868b7c937a",
                }
            }
        })

    def modify_cluster(self):
        request_kwargs = {
            "cluster_identifier": self._get_param('ClusterIdentifier'),
            "new_cluster_identifier": self._get_param('NewClusterIdentifier'),
            "node_type": self._get_param('NodeType'),
            "master_user_password": self._get_param('MasterUserPassword'),
            "cluster_type": self._get_param('ClusterType'),
            "cluster_security_groups": self._get_cluster_security_groups(),
            "vpc_security_group_ids": self._get_vpc_security_group_ids(),
            "cluster_subnet_group_name": self._get_param('ClusterSubnetGroupName'),
            "preferred_maintenance_window": self._get_param('PreferredMaintenanceWindow'),
            "cluster_parameter_group_name": self._get_param('ClusterParameterGroupName'),
            "automated_snapshot_retention_period": self._get_int_param('AutomatedSnapshotRetentionPeriod'),
            "cluster_version": self._get_param('ClusterVersion'),
            "allow_version_upgrade": self._get_bool_param('AllowVersionUpgrade'),
            "number_of_nodes": self._get_int_param('NumberOfNodes'),
            "publicly_accessible": self._get_param("PubliclyAccessible"),
            "encrypted": self._get_param("Encrypted"),
        }
        cluster_kwargs = {}
        # We only want parameters that were actually passed in, otherwise
        # we'll stomp all over our cluster metadata with None values.
        for (key, value) in iteritems(request_kwargs):
            if value is not None and value != []:
                cluster_kwargs[key] = value

        cluster = self.redshift_backend.modify_cluster(**cluster_kwargs)

        return self.get_response({
            "ModifyClusterResponse": {
                "ModifyClusterResult": {
                    "Cluster": cluster.to_json(),
                },
                "ResponseMetadata": {
                    "RequestId": "384ac68d-3775-11df-8963-01868b7c937a",
                }
            }
        })

    def delete_cluster(self):
        cluster_identifier = self._get_param("ClusterIdentifier")
        cluster = self.redshift_backend.delete_cluster(cluster_identifier)

        return self.get_response({
            "DeleteClusterResponse": {
                "DeleteClusterResult": {
                    "Cluster": cluster.to_json()
                },
                "ResponseMetadata": {
                    "RequestId": "384ac68d-3775-11df-8963-01868b7c937a",
                }
            }
        })

    def create_cluster_subnet_group(self):
        cluster_subnet_group_name = self._get_param('ClusterSubnetGroupName')
        description = self._get_param('Description')
        subnet_ids = self._get_subnet_ids()
        tags = self.unpack_complex_list_params('Tags.Tag', ('Key', 'Value'))

        subnet_group = self.redshift_backend.create_cluster_subnet_group(
            cluster_subnet_group_name=cluster_subnet_group_name,
            description=description,
            subnet_ids=subnet_ids,
            region_name=self.region,
            tags=tags
        )

        return self.get_response({
            "CreateClusterSubnetGroupResponse": {
                "CreateClusterSubnetGroupResult": {
                    "ClusterSubnetGroup": subnet_group.to_json(),
                },
                "ResponseMetadata": {
                    "RequestId": "384ac68d-3775-11df-8963-01868b7c937a",
                }
            }
        })

    def describe_cluster_subnet_groups(self):
        subnet_identifier = self._get_param("ClusterSubnetGroupName")
        subnet_groups = self.redshift_backend.describe_cluster_subnet_groups(
            subnet_identifier)

        return self.get_response({
            "DescribeClusterSubnetGroupsResponse": {
                "DescribeClusterSubnetGroupsResult": {
                    "ClusterSubnetGroups": [subnet_group.to_json() for subnet_group in subnet_groups]
                },
                "ResponseMetadata": {
                    "RequestId": "384ac68d-3775-11df-8963-01868b7c937a",
                }
            }
        })

    def delete_cluster_subnet_group(self):
        subnet_identifier = self._get_param("ClusterSubnetGroupName")
        self.redshift_backend.delete_cluster_subnet_group(subnet_identifier)

        return self.get_response({
            "DeleteClusterSubnetGroupResponse": {
                "ResponseMetadata": {
                    "RequestId": "384ac68d-3775-11df-8963-01868b7c937a",
                }
            }
        })

    def create_cluster_security_group(self):
        cluster_security_group_name = self._get_param(
            'ClusterSecurityGroupName')
        description = self._get_param('Description')
        tags = self.unpack_complex_list_params('Tags.Tag', ('Key', 'Value'))

        security_group = self.redshift_backend.create_cluster_security_group(
            cluster_security_group_name=cluster_security_group_name,
            description=description,
            region_name=self.region,
            tags=tags
        )

        return self.get_response({
            "CreateClusterSecurityGroupResponse": {
                "CreateClusterSecurityGroupResult": {
                    "ClusterSecurityGroup": security_group.to_json(),
                },
                "ResponseMetadata": {
                    "RequestId": "384ac68d-3775-11df-8963-01868b7c937a",
                }
            }
        })

    def describe_cluster_security_groups(self):
        cluster_security_group_name = self._get_param(
            "ClusterSecurityGroupName")
        security_groups = self.redshift_backend.describe_cluster_security_groups(
            cluster_security_group_name)

        return self.get_response({
            "DescribeClusterSecurityGroupsResponse": {
                "DescribeClusterSecurityGroupsResult": {
                    "ClusterSecurityGroups": [security_group.to_json() for security_group in security_groups]
                },
                "ResponseMetadata": {
                    "RequestId": "384ac68d-3775-11df-8963-01868b7c937a",
                }
            }
        })

    def delete_cluster_security_group(self):
        security_group_identifier = self._get_param("ClusterSecurityGroupName")
        self.redshift_backend.delete_cluster_security_group(
            security_group_identifier)

        return self.get_response({
            "DeleteClusterSecurityGroupResponse": {
                "ResponseMetadata": {
                    "RequestId": "384ac68d-3775-11df-8963-01868b7c937a",
                }
            }
        })

    def create_cluster_parameter_group(self):
        cluster_parameter_group_name = self._get_param('ParameterGroupName')
        group_family = self._get_param('ParameterGroupFamily')
        description = self._get_param('Description')
        tags = self.unpack_complex_list_params('Tags.Tag', ('Key', 'Value'))

        parameter_group = self.redshift_backend.create_cluster_parameter_group(
            cluster_parameter_group_name,
            group_family,
            description,
            self.region,
            tags
        )

        return self.get_response({
            "CreateClusterParameterGroupResponse": {
                "CreateClusterParameterGroupResult": {
                    "ClusterParameterGroup": parameter_group.to_json(),
                },
                "ResponseMetadata": {
                    "RequestId": "384ac68d-3775-11df-8963-01868b7c937a",
                }
            }
        })

    def describe_cluster_parameter_groups(self):
        cluster_parameter_group_name = self._get_param("ParameterGroupName")
        parameter_groups = self.redshift_backend.describe_cluster_parameter_groups(
            cluster_parameter_group_name)

        return self.get_response({
            "DescribeClusterParameterGroupsResponse": {
                "DescribeClusterParameterGroupsResult": {
                    "ParameterGroups": [parameter_group.to_json() for parameter_group in parameter_groups]
                },
                "ResponseMetadata": {
                    "RequestId": "384ac68d-3775-11df-8963-01868b7c937a",
                }
            }
        })

    def delete_cluster_parameter_group(self):
        cluster_parameter_group_name = self._get_param("ParameterGroupName")
        self.redshift_backend.delete_cluster_parameter_group(
            cluster_parameter_group_name)

        return self.get_response({
            "DeleteClusterParameterGroupResponse": {
                "ResponseMetadata": {
                    "RequestId": "384ac68d-3775-11df-8963-01868b7c937a",
                }
            }
        })

    def create_cluster_snapshot(self):
        cluster_identifier = self._get_param('ClusterIdentifier')
        snapshot_identifier = self._get_param('SnapshotIdentifier')
        tags = self.unpack_complex_list_params('Tags.Tag', ('Key', 'Value'))

        snapshot = self.redshift_backend.create_cluster_snapshot(cluster_identifier,
                                                                 snapshot_identifier,
                                                                 self.region,
                                                                 tags)
        return self.get_response({
            'CreateClusterSnapshotResponse': {
                "CreateClusterSnapshotResult": {
                    "Snapshot": snapshot.to_json(),
                },
                "ResponseMetadata": {
                    "RequestId": "384ac68d-3775-11df-8963-01868b7c937a",
                }
            }
        })

    def describe_cluster_snapshots(self):
        cluster_identifier = self._get_param('ClusterIdentifier')
        snapshot_identifier = self._get_param('SnapshotIdentifier')
        snapshots = self.redshift_backend.describe_cluster_snapshots(cluster_identifier,
                                                                     snapshot_identifier)
        return self.get_response({
            "DescribeClusterSnapshotsResponse": {
                "DescribeClusterSnapshotsResult": {
                    "Snapshots": [snapshot.to_json() for snapshot in snapshots]
                },
                "ResponseMetadata": {
                    "RequestId": "384ac68d-3775-11df-8963-01868b7c937a",
                }
            }
        })

    def delete_cluster_snapshot(self):
        snapshot_identifier = self._get_param('SnapshotIdentifier')
        snapshot = self.redshift_backend.delete_cluster_snapshot(snapshot_identifier)

        return self.get_response({
            "DeleteClusterSnapshotResponse": {
                "DeleteClusterSnapshotResult": {
                    "Snapshot": snapshot.to_json()
                },
                "ResponseMetadata": {
                    "RequestId": "384ac68d-3775-11df-8963-01868b7c937a",
                }
            }
        })

    def create_tags(self):
        resource_name = self._get_param('ResourceName')
        tags = self.unpack_complex_list_params('Tags.Tag', ('Key', 'Value'))

        self.redshift_backend.create_tags(resource_name, tags)

        return self.get_response({
            "CreateTagsResponse": {
                "ResponseMetadata": {
                    "RequestId": "384ac68d-3775-11df-8963-01868b7c937a",
                }
            }
        })

    def describe_tags(self):
        resource_name = self._get_param('ResourceName')
        resource_type = self._get_param('ResourceType')

        tagged_resources = self.redshift_backend.describe_tags(resource_name,
                                                               resource_type)
        return self.get_response({
            "DescribeTagsResponse": {
                "DescribeTagsResult": {
                    "TaggedResources": tagged_resources
                },
                "ResponseMetadata": {
                    "RequestId": "384ac68d-3775-11df-8963-01868b7c937a",
                }
            }
        })

    def delete_tags(self):
        resource_name = self._get_param('ResourceName')
        tag_keys = self.unpack_list_params('TagKeys.TagKey')

        self.redshift_backend.delete_tags(resource_name, tag_keys)

        return self.get_response({
            "DeleteTagsResponse": {
                "ResponseMetadata": {
                    "RequestId": "384ac68d-3775-11df-8963-01868b7c937a",
                }
            }
        })



from __future__ import unicode_literals
from moto.core.responses import BaseResponse
from .models import elbv2_backends
from .exceptions import DuplicateTagKeysError
from .exceptions import LoadBalancerNotFoundError
from .exceptions import TargetGroupNotFoundError


class ELBV2Response(BaseResponse):

    @property
    def elbv2_backend(self):
        return elbv2_backends[self.region]

    def create_load_balancer(self):
        load_balancer_name = self._get_param('Name')
        subnet_ids = self._get_multi_param("Subnets.member")
        security_groups = self._get_multi_param("SecurityGroups.member")
        scheme = self._get_param('Scheme')

        load_balancer = self.elbv2_backend.create_load_balancer(
            name=load_balancer_name,
            security_groups=security_groups,
            subnet_ids=subnet_ids,
            scheme=scheme,
        )
        self._add_tags(load_balancer)
        template = self.response_template(CREATE_LOAD_BALANCER_TEMPLATE)
        return template.render(load_balancer=load_balancer)

    def create_rule(self):
        lister_arn = self._get_param('ListenerArn')
        _conditions = self._get_list_prefix('Conditions.member')
        conditions = []
        for _condition in _conditions:
            condition = {}
            condition['field'] = _condition['field']
            values = sorted(
                [e for e in _condition.items() if e[0].startswith('values.member')],
                key=lambda x: x[0]
            )
            condition['values'] = [e[1] for e in values]
            conditions.append(condition)
        priority = self._get_int_param('Priority')
        actions = self._get_list_prefix('Actions.member')
        rules = self.elbv2_backend.create_rule(
            listener_arn=lister_arn,
            conditions=conditions,
            priority=priority,
            actions=actions
        )
        template = self.response_template(CREATE_RULE_TEMPLATE)
        return template.render(rules=rules)

    def create_target_group(self):
        name = self._get_param('Name')
        vpc_id = self._get_param('VpcId')
        protocol = self._get_param('Protocol')
        port = self._get_param('Port')
        healthcheck_protocol = self._get_param('HealthCheckProtocol', 'HTTP')
        healthcheck_port = self._get_param('HealthCheckPort', 'traffic-port')
        healthcheck_path = self._get_param('HealthCheckPath', '/')
        healthcheck_interval_seconds = self._get_param('HealthCheckIntervalSeconds', '30')
        healthcheck_timeout_seconds = self._get_param('HealthCheckTimeoutSeconds', '5')
        healthy_threshold_count = self._get_param('HealthyThresholdCount', '5')
        unhealthy_threshold_count = self._get_param('UnhealthyThresholdCount', '2')

        target_group = self.elbv2_backend.create_target_group(
            name,
            vpc_id=vpc_id,
            protocol=protocol,
            port=port,
            healthcheck_protocol=healthcheck_protocol,
            healthcheck_port=healthcheck_port,
            healthcheck_path=healthcheck_path,
            healthcheck_interval_seconds=healthcheck_interval_seconds,
            healthcheck_timeout_seconds=healthcheck_timeout_seconds,
            healthy_threshold_count=healthy_threshold_count,
            unhealthy_threshold_count=unhealthy_threshold_count,
        )

        template = self.response_template(CREATE_TARGET_GROUP_TEMPLATE)
        return template.render(target_group=target_group)

    def create_listener(self):
        load_balancer_arn = self._get_param('LoadBalancerArn')
        protocol = self._get_param('Protocol')
        port = self._get_param('Port')
        ssl_policy = self._get_param('SslPolicy', 'ELBSecurityPolicy-2016-08')
        certificates = self._get_list_prefix('Certificates.member')
        if certificates:
            certificate = certificates[0].get('certificate_arn')
        else:
            certificate = None
        default_actions = self._get_list_prefix('DefaultActions.member')

        listener = self.elbv2_backend.create_listener(
            load_balancer_arn=load_balancer_arn,
            protocol=protocol,
            port=port,
            ssl_policy=ssl_policy,
            certificate=certificate,
            default_actions=default_actions)

        template = self.response_template(CREATE_LISTENER_TEMPLATE)
        return template.render(listener=listener)

    def describe_load_balancers(self):
        arns = self._get_multi_param("LoadBalancerArns.member")
        names = self._get_multi_param("Names.member")
        all_load_balancers = list(self.elbv2_backend.describe_load_balancers(arns, names))
        marker = self._get_param('Marker')
        all_names = [balancer.name for balancer in all_load_balancers]
        if marker:
            start = all_names.index(marker) + 1
        else:
            start = 0
        page_size = self._get_param('PageSize', 50)  # the default is 400, but using 50 to make testing easier
        load_balancers_resp = all_load_balancers[start:start + page_size]
        next_marker = None
        if len(all_load_balancers) > start + page_size:
            next_marker = load_balancers_resp[-1].name

        template = self.response_template(DESCRIBE_LOAD_BALANCERS_TEMPLATE)
        return template.render(load_balancers=load_balancers_resp, marker=next_marker)

    def describe_rules(self):
        listener_arn = self._get_param('ListenerArn')
        rule_arns = self._get_multi_param('RuleArns.member') if any(k for k in list(self.querystring.keys()) if k.startswith('RuleArns.member')) else None
        all_rules = self.elbv2_backend.describe_rules(listener_arn, rule_arns)
        all_arns = [rule.arn for rule in all_rules]
        page_size = self._get_int_param('PageSize', 50)  # set 50 for temporary

        marker = self._get_param('Marker')
        if marker:
            start = all_arns.index(marker) + 1
        else:
            start = 0
        rules_resp = all_rules[start:start + page_size]
        next_marker = None

        if len(all_rules) > start + page_size:
            next_marker = rules_resp[-1].arn
        template = self.response_template(DESCRIBE_RULES_TEMPLATE)
        return template.render(rules=rules_resp, marker=next_marker)

    def describe_target_groups(self):
        load_balancer_arn = self._get_param('LoadBalancerArn')
        target_group_arns = self._get_multi_param('TargetGroupArns.member')
        names = self._get_multi_param('Names.member')

        target_groups = self.elbv2_backend.describe_target_groups(load_balancer_arn, target_group_arns, names)
        template = self.response_template(DESCRIBE_TARGET_GROUPS_TEMPLATE)
        return template.render(target_groups=target_groups)

    def describe_target_group_attributes(self):
        target_group_arn = self._get_param('TargetGroupArn')
        target_group = self.elbv2_backend.target_groups.get(target_group_arn)
        if not target_group:
            raise TargetGroupNotFoundError()
        template = self.response_template(DESCRIBE_TARGET_GROUP_ATTRIBUTES_TEMPLATE)
        return template.render(attributes=target_group.attributes)

    def describe_listeners(self):
        load_balancer_arn = self._get_param('LoadBalancerArn')
        listener_arns = self._get_multi_param('ListenerArns.member')
        if not load_balancer_arn and not listener_arns:
            raise LoadBalancerNotFoundError()

        listeners = self.elbv2_backend.describe_listeners(load_balancer_arn, listener_arns)
        template = self.response_template(DESCRIBE_LISTENERS_TEMPLATE)
        return template.render(listeners=listeners)

    def delete_load_balancer(self):
        arn = self._get_param('LoadBalancerArn')
        self.elbv2_backend.delete_load_balancer(arn)
        template = self.response_template(DELETE_LOAD_BALANCER_TEMPLATE)
        return template.render()

    def delete_rule(self):
        arn = self._get_param('RuleArn')
        self.elbv2_backend.delete_rule(arn)
        template = self.response_template(DELETE_RULE_TEMPLATE)
        return template.render()

    def delete_target_group(self):
        arn = self._get_param('TargetGroupArn')
        self.elbv2_backend.delete_target_group(arn)
        template = self.response_template(DELETE_TARGET_GROUP_TEMPLATE)
        return template.render()

    def delete_listener(self):
        arn = self._get_param('ListenerArn')
        self.elbv2_backend.delete_listener(arn)
        template = self.response_template(DELETE_LISTENER_TEMPLATE)
        return template.render()

    def modify_rule(self):
        rule_arn = self._get_param('RuleArn')
        _conditions = self._get_list_prefix('Conditions.member')
        conditions = []
        for _condition in _conditions:
            condition = {}
            condition['field'] = _condition['field']
            values = sorted(
                [e for e in _condition.items() if e[0].startswith('values.member')],
                key=lambda x: x[0]
            )
            condition['values'] = [e[1] for e in values]
            conditions.append(condition)
        actions = self._get_list_prefix('Actions.member')
        rules = self.elbv2_backend.modify_rule(
            rule_arn=rule_arn,
            conditions=conditions,
            actions=actions
        )
        template = self.response_template(MODIFY_RULE_TEMPLATE)
        return template.render(rules=rules)

    def modify_target_group_attributes(self):
        target_group_arn = self._get_param('TargetGroupArn')
        target_group = self.elbv2_backend.target_groups.get(target_group_arn)
        attributes = {
            attr['key']: attr['value']
            for attr in self._get_list_prefix('Attributes.member')
        }
        target_group.attributes.update(attributes)
        if not target_group:
            raise TargetGroupNotFoundError()
        template = self.response_template(MODIFY_TARGET_GROUP_ATTRIBUTES_TEMPLATE)
        return template.render(attributes=attributes)

    def register_targets(self):
        target_group_arn = self._get_param('TargetGroupArn')
        targets = self._get_list_prefix('Targets.member')
        self.elbv2_backend.register_targets(target_group_arn, targets)

        template = self.response_template(REGISTER_TARGETS_TEMPLATE)
        return template.render()

    def deregister_targets(self):
        target_group_arn = self._get_param('TargetGroupArn')
        targets = self._get_list_prefix('Targets.member')
        self.elbv2_backend.deregister_targets(target_group_arn, targets)

        template = self.response_template(DEREGISTER_TARGETS_TEMPLATE)
        return template.render()

    def describe_target_health(self):
        target_group_arn = self._get_param('TargetGroupArn')
        targets = self._get_list_prefix('Targets.member')
        target_health_descriptions = self.elbv2_backend.describe_target_health(target_group_arn, targets)

        template = self.response_template(DESCRIBE_TARGET_HEALTH_TEMPLATE)
        return template.render(target_health_descriptions=target_health_descriptions)

    def set_rule_priorities(self):
        rule_priorities = self._get_list_prefix('RulePriorities.member')
        for rule_priority in rule_priorities:
            rule_priority['priority'] = int(rule_priority['priority'])
        rules = self.elbv2_backend.set_rule_priorities(rule_priorities)
        template = self.response_template(SET_RULE_PRIORITIES_TEMPLATE)
        return template.render(rules=rules)

    def add_tags(self):
        resource_arns = self._get_multi_param('ResourceArns.member')

        for arn in resource_arns:
            if ':targetgroup' in arn:
                resource = self.elbv2_backend.target_groups.get(arn)
                if not resource:
                    raise TargetGroupNotFoundError()
            elif ':loadbalancer' in arn:
                resource = self.elbv2_backend.load_balancers.get(arn)
                if not resource:
                    raise LoadBalancerNotFoundError()
            else:
                raise LoadBalancerNotFoundError()
            self._add_tags(resource)

        template = self.response_template(ADD_TAGS_TEMPLATE)
        return template.render()

    def remove_tags(self):
        resource_arns = self._get_multi_param('ResourceArns.member')
        tag_keys = self._get_multi_param('TagKeys.member')

        for arn in resource_arns:
            if ':targetgroup' in arn:
                resource = self.elbv2_backend.target_groups.get(arn)
                if not resource:
                    raise TargetGroupNotFoundError()
            elif ':loadbalancer' in arn:
                resource = self.elbv2_backend.load_balancers.get(arn)
                if not resource:
                    raise LoadBalancerNotFoundError()
            else:
                raise LoadBalancerNotFoundError()
            [resource.remove_tag(key) for key in tag_keys]

        template = self.response_template(REMOVE_TAGS_TEMPLATE)
        return template.render()

    def describe_tags(self):
        resource_arns = self._get_multi_param('ResourceArns.member')
        resources = []
        for arn in resource_arns:
            if ':targetgroup' in arn:
                resource = self.elbv2_backend.target_groups.get(arn)
                if not resource:
                    raise TargetGroupNotFoundError()
            elif ':loadbalancer' in arn:
                resource = self.elbv2_backend.load_balancers.get(arn)
                if not resource:
                    raise LoadBalancerNotFoundError()
            else:
                raise LoadBalancerNotFoundError()
            resources.append(resource)

        template = self.response_template(DESCRIBE_TAGS_TEMPLATE)
        return template.render(resources=resources)

    def _add_tags(self, resource):
        tag_values = []
        tag_keys = []

        for t_key, t_val in sorted(self.querystring.items()):
            if t_key.startswith('Tags.member.'):
                if t_key.split('.')[3] == 'Key':
                    tag_keys.extend(t_val)
                elif t_key.split('.')[3] == 'Value':
                    tag_values.extend(t_val)

        counts = {}
        for i in tag_keys:
            counts[i] = tag_keys.count(i)

        counts = sorted(counts.items(), key=lambda i: i[1], reverse=True)

        if counts and counts[0][1] > 1:
            # We have dupes...
            raise DuplicateTagKeysError(counts[0])

        for tag_key, tag_value in zip(tag_keys, tag_values):
            resource.add_tag(tag_key, tag_value)


ADD_TAGS_TEMPLATE = """<AddTagsResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <AddTagsResult/>
  <ResponseMetadata>
    <RequestId>360e81f7-1100-11e4-b6ed-0f30EXAMPLE</RequestId>
  </ResponseMetadata>
</AddTagsResponse>"""

REMOVE_TAGS_TEMPLATE = """<RemoveTagsResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <RemoveTagsResult/>
  <ResponseMetadata>
    <RequestId>360e81f7-1100-11e4-b6ed-0f30EXAMPLE</RequestId>
  </ResponseMetadata>
</RemoveTagsResponse>"""

DESCRIBE_TAGS_TEMPLATE = """<DescribeTagsResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <DescribeTagsResult>
    <TagDescriptions>
      {% for resource in resources %}
      <member>
        <ResourceArn>{{ resource.arn }}</ResourceArn>
        <Tags>
          {% for key, value in resource.tags.items() %}
          <member>
            <Value>{{ value }}</Value>
            <Key>{{ key }}</Key>
          </member>
          {% endfor %}
        </Tags>
      </member>
      {% endfor %}
    </TagDescriptions>
  </DescribeTagsResult>
  <ResponseMetadata>
    <RequestId>360e81f7-1100-11e4-b6ed-0f30EXAMPLE</RequestId>
  </ResponseMetadata>
</DescribeTagsResponse>"""


CREATE_LOAD_BALANCER_TEMPLATE = """<CreateLoadBalancerResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <CreateLoadBalancerResult>
    <LoadBalancers>
      <member>
        <LoadBalancerArn>{{ load_balancer.arn }}</LoadBalancerArn>
        <Scheme>{{ load_balancer.scheme }}</Scheme>
        <LoadBalancerName>{{ load_balancer.name }}</LoadBalancerName>
        <VpcId>{{ load_balancer.vpc_id }}</VpcId>
        <CanonicalHostedZoneId>Z2P70J7EXAMPLE</CanonicalHostedZoneId>
        <CreatedTime>{{ load_balancer.created_time }}</CreatedTime>
        <AvailabilityZones>
          {% for subnet in load_balancer.subnets %}
          <member>
            <SubnetId>{{ subnet.id }}</SubnetId>
            <ZoneName>{{ subnet.availability_zone }}</ZoneName>
          </member>
          {% endfor %}
        </AvailabilityZones>
        <SecurityGroups>
          {% for security_group in load_balancer.security_groups %}
          <member>{{ security_group }}</member>
          {% endfor %}
        </SecurityGroups>
        <DNSName>{{ load_balancer.dns_name }}</DNSName>
        <State>
          <Code>provisioning</Code>
        </State>
        <Type>application</Type>
      </member>
    </LoadBalancers>
  </CreateLoadBalancerResult>
  <ResponseMetadata>
    <RequestId>32d531b2-f2d0-11e5-9192-3fff33344cfa</RequestId>
  </ResponseMetadata>
</CreateLoadBalancerResponse>"""

CREATE_RULE_TEMPLATE = """<CreateRuleResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <CreateRuleResult>
    <Rules>
      {% for rule in rules %}
      <member>
        <IsDefault>{{ "true" if rule.is_default else "false" }}</IsDefault>
        <Conditions>
          {% for condition in rule.conditions %}
          <member>
            <Field>{{ condition["field"] }}</Field>
            <Values>
              {% for value in condition["values"] %}
              <member>{{ value }}</member>
              {% endfor %}
            </Values>
          </member>
          {% endfor %}
        </Conditions>
        <Priority>{{ rule.priority }}</Priority>
        <Actions>
          {% for action in rule.actions %}
          <member>
            <Type>{{ action["type"] }}</Type>
            <TargetGroupArn>{{ action["target_group_arn"] }}</TargetGroupArn>
          </member>
          {% endfor %}
        </Actions>
        <RuleArn>{{ rule.arn }}</RuleArn>
      </member>
      {% endfor %}
    </Rules>
  </CreateRuleResult>
  <ResponseMetadata>
    <RequestId>c5478c83-f397-11e5-bb98-57195a6eb84a</RequestId>
  </ResponseMetadata>
</CreateRuleResponse>"""

CREATE_TARGET_GROUP_TEMPLATE = """<CreateTargetGroupResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <CreateTargetGroupResult>
    <TargetGroups>
      <member>
        <TargetGroupArn>{{ target_group.arn }}</TargetGroupArn>
        <TargetGroupName>{{ target_group.name }}</TargetGroupName>
        <Protocol>{{ target_group.protocol }}</Protocol>
        <Port>{{ target_group.port }}</Port>
        <VpcId>{{ target_group.vpc_id }}</VpcId>
        <HealthCheckProtocol>{{ target_group.health_check_protocol }}</HealthCheckProtocol>
        <HealthCheckPort>{{ target_group.healthcheck_port }}</HealthCheckPort>
        <HealthCheckPath>{{ target_group.healthcheck_path }}</HealthCheckPath>
        <HealthCheckIntervalSeconds>{{ target_group.healthcheck_interval_seconds }}</HealthCheckIntervalSeconds>
        <HealthCheckTimeoutSeconds>{{ target_group.healthcheck_timeout_seconds }}</HealthCheckTimeoutSeconds>
        <HealthyThresholdCount>{{ target_group.healthy_threshold_count }}</HealthyThresholdCount>
        <UnhealthyThresholdCount>{{ target_group.unhealthy_threshold_count }}</UnhealthyThresholdCount>
        <Matcher>
          <HttpCode>200</HttpCode>
        </Matcher>
      </member>
    </TargetGroups>
  </CreateTargetGroupResult>
  <ResponseMetadata>
    <RequestId>b83fe90e-f2d5-11e5-b95d-3b2c1831fc26</RequestId>
  </ResponseMetadata>
</CreateTargetGroupResponse>"""

CREATE_LISTENER_TEMPLATE = """<CreateListenerResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <CreateListenerResult>
    <Listeners>
      <member>
        <LoadBalancerArn>{{ listener.load_balancer_arn }}</LoadBalancerArn>
        <Protocol>{{ listener.protocol }}</Protocol>
        {% if listener.certificate %}
        <Certificates>
          <member>
            <CertificateArn>{{ listener.certificate }}</CertificateArn>
          </member>
        </Certificates>
        {% endif %}
        <Port>{{ listener.port }}</Port>
        <SslPolicy>{{ listener.ssl_policy }}</SslPolicy>
        <ListenerArn>{{ listener.arn }}</ListenerArn>
        <DefaultActions>
          {% for action in listener.default_actions %}
          <member>
            <Type>{{ action.type }}</Type>
            <TargetGroupArn>{{ action.target_group_arn }}</TargetGroupArn>
          </member>
          {% endfor %}
        </DefaultActions>
      </member>
    </Listeners>
  </CreateListenerResult>
  <ResponseMetadata>
    <RequestId>97f1bb38-f390-11e5-b95d-3b2c1831fc26</RequestId>
  </ResponseMetadata>
</CreateListenerResponse>"""

DELETE_LOAD_BALANCER_TEMPLATE = """<DeleteLoadBalancerResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <DeleteLoadBalancerResult/>
  <ResponseMetadata>
    <RequestId>1549581b-12b7-11e3-895e-1334aEXAMPLE</RequestId>
  </ResponseMetadata>
</DeleteLoadBalancerResponse>"""

DELETE_RULE_TEMPLATE = """<DeleteRuleResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <DeleteRuleResult/>
  <ResponseMetadata>
    <RequestId>1549581b-12b7-11e3-895e-1334aEXAMPLE</RequestId>
  </ResponseMetadata>
</DeleteRuleResponse>"""

DELETE_TARGET_GROUP_TEMPLATE = """<DeleteTargetGroupResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <DeleteTargetGroupResult/>
  <ResponseMetadata>
    <RequestId>1549581b-12b7-11e3-895e-1334aEXAMPLE</RequestId>
  </ResponseMetadata>
</DeleteTargetGroupResponse>"""

DELETE_LISTENER_TEMPLATE = """<DeleteListenerResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <DeleteListenerResult/>
  <ResponseMetadata>
    <RequestId>1549581b-12b7-11e3-895e-1334aEXAMPLE</RequestId>
  </ResponseMetadata>
</DeleteListenerResponse>"""

DESCRIBE_LOAD_BALANCERS_TEMPLATE = """<DescribeLoadBalancersResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <DescribeLoadBalancersResult>
    <LoadBalancers>
      {% for load_balancer in load_balancers %}
      <member>
        <LoadBalancerArn>{{ load_balancer.arn }}</LoadBalancerArn>
        <Scheme>{{ load_balancer.scheme }}</Scheme>
        <LoadBalancerName>{{ load_balancer.name }}</LoadBalancerName>
        <VpcId>{{ load_balancer.vpc_id }}</VpcId>
        <CanonicalHostedZoneId>Z2P70J7EXAMPLE</CanonicalHostedZoneId>
        <CreatedTime>{{ load_balancer.created_time }}</CreatedTime>
        <AvailabilityZones>
          {% for subnet in load_balancer.subnets %}
          <member>
            <SubnetId>{{ subnet.id }}</SubnetId>
            <ZoneName>{{ subnet.availability_zone }}</ZoneName>
          </member>
          {% endfor %}
        </AvailabilityZones>
        <SecurityGroups>
          {% for security_group in load_balancer.security_groups %}
          <member>{{ security_group }}</member>
          {% endfor %}
        </SecurityGroups>
        <DNSName>{{ load_balancer.dns_name }}</DNSName>
        <State>
          <Code>provisioning</Code>
        </State>
        <Type>application</Type>
      </member>
      {% endfor %}
    </LoadBalancers>
    {% if marker %}
    <NextMarker>{{ marker }}</NextMarker>
    {% endif %}
  </DescribeLoadBalancersResult>
  <ResponseMetadata>
    <RequestId>f9880f01-7852-629d-a6c3-3ae2-666a409287e6dc0c</RequestId>
  </ResponseMetadata>
</DescribeLoadBalancersResponse>"""

DESCRIBE_RULES_TEMPLATE = """<DescribeRulesResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <DescribeRulesResult>
    <Rules>
      {% for rule in rules %}
      <member>
        <IsDefault>{{ "true" if rule.is_default else "false" }}</IsDefault>
        <Conditions>
          {% for condition in rule.conditions %}
          <member>
            <Field>{{ condition["field"] }}</Field>
            <Values>
              {% for value in condition["values"] %}
              <member>{{ value }}</member>
              {% endfor %}
            </Values>
          </member>
          {% endfor %}
        </Conditions>
        <Priority>{{ rule.priority }}</Priority>
        <Actions>
          {% for action in rule.actions %}
          <member>
            <Type>{{ action["type"] }}</Type>
            <TargetGroupArn>{{ action["target_group_arn"] }}</TargetGroupArn>
          </member>
          {% endfor %}
        </Actions>
        <RuleArn>{{ rule.arn }}</RuleArn>
      </member>
      {% endfor %}
    </Rules>
    {% if marker %}
    <NextMarker>{{ marker }}</NextMarker>
    {% endif %}
  </DescribeRulesResult>
  <ResponseMetadata>
    <RequestId>74926cf3-f3a3-11e5-b543-9f2c3fbb9bee</RequestId>
  </ResponseMetadata>
</DescribeRulesResponse>"""

DESCRIBE_TARGET_GROUPS_TEMPLATE = """<DescribeTargetGroupsResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <DescribeTargetGroupsResult>
    <TargetGroups>
      {% for target_group in target_groups %}
      <member>
        <TargetGroupArn>{{ target_group.arn }}</TargetGroupArn>
        <TargetGroupName>{{ target_group.name }}</TargetGroupName>
        <Protocol>{{ target_group.protocol }}</Protocol>
        <Port>{{ target_group.port }}</Port>
        <VpcId>{{ target_group.vpc_id }}</VpcId>
        <HealthCheckProtocol>{{ target_group.health_check_protocol }}</HealthCheckProtocol>
        <HealthCheckPort>{{ target_group.healthcheck_port }}</HealthCheckPort>
        <HealthCheckPath>{{ target_group.healthcheck_path }}</HealthCheckPath>
        <HealthCheckIntervalSeconds>{{ target_group.healthcheck_interval_seconds }}</HealthCheckIntervalSeconds>
        <HealthCheckTimeoutSeconds>{{ target_group.healthcheck_timeout_seconds }}</HealthCheckTimeoutSeconds>
        <HealthyThresholdCount>{{ target_group.healthy_threshold_count }}</HealthyThresholdCount>
        <UnhealthyThresholdCount>{{ target_group.unhealthy_threshold_count }}</UnhealthyThresholdCount>
        <Matcher>
          <HttpCode>200</HttpCode>
        </Matcher>
        <LoadBalancerArns>
          {% for load_balancer_arn in target_group.load_balancer_arns %}
          <member>{{ load_balancer_arn }}</member>
          {% endfor %}
        </LoadBalancerArns>
      </member>
      {% endfor %}
    </TargetGroups>
  </DescribeTargetGroupsResult>
  <ResponseMetadata>
    <RequestId>70092c0e-f3a9-11e5-ae48-cff02092876b</RequestId>
  </ResponseMetadata>
</DescribeTargetGroupsResponse>"""


DESCRIBE_TARGET_GROUP_ATTRIBUTES_TEMPLATE = """<DescribeTargetGroupAttributesResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <DescribeTargetGroupAttributesResult>
    <Attributes>
      {% for key, value in attributes.items() %}
      <member>
        <Key>{{ key }}</Key>
        <Value>{{ value }}</Value>
      </member>
      {% endfor %}
    </Attributes>
  </DescribeTargetGroupAttributesResult>
  <ResponseMetadata>
    <RequestId>70092c0e-f3a9-11e5-ae48-cff02092876b</RequestId>
  </ResponseMetadata>
</DescribeTargetGroupAttributesResponse>"""


DESCRIBE_LISTENERS_TEMPLATE = """<DescribeLoadBalancersResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <DescribeListenersResult>
    <Listeners>
      {% for listener in listeners %}
      <member>
        <LoadBalancerArn>{{ listener.load_balancer_arn }}</LoadBalancerArn>
        <Protocol>{{ listener.protocol }}</Protocol>
        {% if listener.certificate %}
        <Certificates>
          <member>
            <CertificateArn>{{ listener.certificate }}</CertificateArn>
          </member>
        </Certificates>
        {% endif %}
        <Port>{{ listener.port }}</Port>
        <SslPolicy>{{ listener.ssl_policy }}</SslPolicy>
        <ListenerArn>{{ listener.arn }}</ListenerArn>
        <DefaultActions>
          {% for action in listener.default_actions %}
          <member>
            <Type>{{ action.type }}</Type>
            <TargetGroupArn>{{ action.target_group_arn }}</TargetGroupArn>
          </member>
          {% endfor %}
        </DefaultActions>
      </member>
      {% endfor %}
    </Listeners>
  </DescribeListenersResult>
  <ResponseMetadata>
    <RequestId>65a3a7ea-f39c-11e5-b543-9f2c3fbb9bee</RequestId>
  </ResponseMetadata>
</DescribeLoadBalancersResponse>"""

CONFIGURE_HEALTH_CHECK_TEMPLATE = """<ConfigureHealthCheckResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <ConfigureHealthCheckResult>
    <HealthCheck>
      <Interval>{{ check.interval }}</Interval>
      <Target>{{ check.target }}</Target>
      <HealthyThreshold>{{ check.healthy_threshold }}</HealthyThreshold>
      <Timeout>{{ check.timeout }}</Timeout>
      <UnhealthyThreshold>{{ check.unhealthy_threshold }}</UnhealthyThreshold>
    </HealthCheck>
  </ConfigureHealthCheckResult>
  <ResponseMetadata>
    <RequestId>f9880f01-7852-629d-a6c3-3ae2-666a409287e6dc0c</RequestId>
  </ResponseMetadata>
</ConfigureHealthCheckResponse>"""

MODIFY_RULE_TEMPLATE = """<ModifyRuleResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <ModifyRuleResult>
    <Rules>
      {% for rule in rules %}
      <member>
        <IsDefault>{{ "true" if rule.is_default else "false" }}</IsDefault>
        <Conditions>
          {% for condition in rule.conditions %}
          <member>
            <Field>{{ condition["field"] }}</Field>
            <Values>
              {% for value in condition["values"] %}
              <member>{{ value }}</member>
              {% endfor %}
            </Values>
          </member>
          {% endfor %}
        </Conditions>
        <Priority>{{ rule.priority }}</Priority>
        <Actions>
          {% for action in rule.actions %}
          <member>
            <Type>{{ action["type"] }}</Type>
            <TargetGroupArn>{{ action["target_group_arn"] }}</TargetGroupArn>
          </member>
          {% endfor %}
        </Actions>
        <RuleArn>{{ rule.arn }}</RuleArn>
      </member>
      {% endfor %}
    </Rules>
  </ModifyRuleResult>
  <ResponseMetadata>
    <RequestId>c5478c83-f397-11e5-bb98-57195a6eb84a</RequestId>
  </ResponseMetadata>
</ModifyRuleResponse>"""

MODIFY_TARGET_GROUP_ATTRIBUTES_TEMPLATE = """<ModifyTargetGroupAttributesResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <ModifyTargetGroupAttributesResult>
    <Attributes>
      {% for key, value in attributes.items() %}
      <member>
        <Key>{{ key }}</Key>
        <Value>{{ value }}</Value>
      </member>
      {% endfor %}
    </Attributes>
  </ModifyTargetGroupAttributesResult>
  <ResponseMetadata>
    <RequestId>70092c0e-f3a9-11e5-ae48-cff02092876b</RequestId>
  </ResponseMetadata>
</ModifyTargetGroupAttributesResponse>"""

REGISTER_TARGETS_TEMPLATE = """<RegisterTargetsResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <RegisterTargetsResult>
  </RegisterTargetsResult>
  <ResponseMetadata>
    <RequestId>f9880f01-7852-629d-a6c3-3ae2-666a409287e6dc0c</RequestId>
  </ResponseMetadata>
</RegisterTargetsResponse>"""

DEREGISTER_TARGETS_TEMPLATE = """<DeregisterTargetsResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <DeregisterTargetsResult>
  </DeregisterTargetsResult>
  <ResponseMetadata>
    <RequestId>f9880f01-7852-629d-a6c3-3ae2-666a409287e6dc0c</RequestId>
  </ResponseMetadata>
</DeregisterTargetsResponse>"""

SET_LOAD_BALANCER_SSL_CERTIFICATE = """<SetLoadBalancerListenerSSLCertificateResponse xmlns="http://elasticloadbalan cing.amazonaws.com/doc/2015-12-01/">
 <SetLoadBalancerListenerSSLCertificateResult/>
<ResponseMetadata>
    <RequestId>83c88b9d-12b7-11e3-8b82-87b12EXAMPLE</RequestId>
</ResponseMetadata>
</SetLoadBalancerListenerSSLCertificateResponse>"""


DELETE_LOAD_BALANCER_LISTENERS = """<DeleteLoadBalancerListenersResponse xmlns="http://elasticloadbalan cing.amazonaws.com/doc/2015-12-01/">
 <DeleteLoadBalancerListenersResult/>
<ResponseMetadata>
    <RequestId>83c88b9d-12b7-11e3-8b82-87b12EXAMPLE</RequestId>
</ResponseMetadata>
</DeleteLoadBalancerListenersResponse>"""

DESCRIBE_ATTRIBUTES_TEMPLATE = """<DescribeLoadBalancerAttributesResponse  xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <DescribeLoadBalancerAttributesResult>
    <LoadBalancerAttributes>
      <AccessLog>
        <Enabled>{{ attributes.access_log.enabled }}</Enabled>
        {% if attributes.access_log.enabled %}
        <S3BucketName>{{ attributes.access_log.s3_bucket_name }}</S3BucketName>
        <S3BucketPrefix>{{ attributes.access_log.s3_bucket_prefix }}</S3BucketPrefix>
        <EmitInterval>{{ attributes.access_log.emit_interval }}</EmitInterval>
        {% endif %}
      </AccessLog>
      <ConnectionSettings>
        <IdleTimeout>{{ attributes.connecting_settings.idle_timeout }}</IdleTimeout>
      </ConnectionSettings>
      <CrossZoneLoadBalancing>
        <Enabled>{{ attributes.cross_zone_load_balancing.enabled }}</Enabled>
      </CrossZoneLoadBalancing>
      <ConnectionDraining>
        {% if attributes.connection_draining.enabled %}
        <Enabled>true</Enabled>
        <Timeout>{{ attributes.connection_draining.timeout }}</Timeout>
        {% else %}
        <Enabled>false</Enabled>
        {% endif %}
      </ConnectionDraining>
    </LoadBalancerAttributes>
  </DescribeLoadBalancerAttributesResult>
  <ResponseMetadata>
    <RequestId>83c88b9d-12b7-11e3-8b82-87b12EXAMPLE</RequestId>
  </ResponseMetadata>
</DescribeLoadBalancerAttributesResponse>
"""

MODIFY_ATTRIBUTES_TEMPLATE = """<ModifyLoadBalancerAttributesResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <ModifyLoadBalancerAttributesResult>
  <LoadBalancerName>{{ load_balancer.name }}</LoadBalancerName>
    <LoadBalancerAttributes>
      <AccessLog>
        <Enabled>{{ attributes.access_log.enabled }}</Enabled>
        {% if attributes.access_log.enabled %}
        <S3BucketName>{{ attributes.access_log.s3_bucket_name }}</S3BucketName>
        <S3BucketPrefix>{{ attributes.access_log.s3_bucket_prefix }}</S3BucketPrefix>
        <EmitInterval>{{ attributes.access_log.emit_interval }}</EmitInterval>
        {% endif %}
      </AccessLog>
      <ConnectionSettings>
        <IdleTimeout>{{ attributes.connecting_settings.idle_timeout }}</IdleTimeout>
      </ConnectionSettings>
      <CrossZoneLoadBalancing>
        <Enabled>{{ attributes.cross_zone_load_balancing.enabled }}</Enabled>
      </CrossZoneLoadBalancing>
      <ConnectionDraining>
        {% if attributes.connection_draining.enabled %}
        <Enabled>true</Enabled>
        <Timeout>{{ attributes.connection_draining.timeout }}</Timeout>
        {% else %}
        <Enabled>false</Enabled>
        {% endif %}
      </ConnectionDraining>
    </LoadBalancerAttributes>
  </ModifyLoadBalancerAttributesResult>
  <ResponseMetadata>
    <RequestId>83c88b9d-12b7-11e3-8b82-87b12EXAMPLE</RequestId>
  </ResponseMetadata>
</ModifyLoadBalancerAttributesResponse>
"""

CREATE_LOAD_BALANCER_POLICY_TEMPLATE = """<CreateLoadBalancerPolicyResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <CreateLoadBalancerPolicyResult/>
  <ResponseMetadata>
      <RequestId>83c88b9d-12b7-11e3-8b82-87b12EXAMPLE</RequestId>
  </ResponseMetadata>
</CreateLoadBalancerPolicyResponse>
"""

SET_LOAD_BALANCER_POLICIES_OF_LISTENER_TEMPLATE = """<SetLoadBalancerPoliciesOfListenerResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
    <SetLoadBalancerPoliciesOfListenerResult/>
    <ResponseMetadata>
        <RequestId>07b1ecbc-1100-11e3-acaf-dd7edEXAMPLE</RequestId>
    </ResponseMetadata>
</SetLoadBalancerPoliciesOfListenerResponse>
"""

SET_LOAD_BALANCER_POLICIES_FOR_BACKEND_SERVER_TEMPLATE = """<SetLoadBalancerPoliciesForBackendServerResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
    <SetLoadBalancerPoliciesForBackendServerResult/>
    <ResponseMetadata>
        <RequestId>0eb9b381-dde0-11e2-8d78-6ddbaEXAMPLE</RequestId>
    </ResponseMetadata>
</SetLoadBalancerPoliciesForBackendServerResponse>
"""

DESCRIBE_TARGET_HEALTH_TEMPLATE = """<DescribeTargetHealthResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <DescribeTargetHealthResult>
    <TargetHealthDescriptions>
      {% for target_health in target_health_descriptions %}
      <member>
        <HealthCheckPort>{{ target_health.health_port }}</HealthCheckPort>
        <TargetHealth>
          <State>{{ target_health.status }}</State>
        </TargetHealth>
        <Target>
          <Port>{{ target_health.port }}</Port>
          <Id>{{ target_health.instance_id }}</Id>
        </Target>
      </member>
      {% endfor %}
    </TargetHealthDescriptions>
  </DescribeTargetHealthResult>
  <ResponseMetadata>
    <RequestId>c534f810-f389-11e5-9192-3fff33344cfa</RequestId>
  </ResponseMetadata>
</DescribeTargetHealthResponse>"""

SET_RULE_PRIORITIES_TEMPLATE = """<SetRulePrioritiesResponse xmlns="http://elasticloadbalancing.amazonaws.com/doc/2015-12-01/">
  <SetRulePrioritiesResult>
    <Rules>
      {% for rule in rules %}
      <member>
        <IsDefault>{{ "true" if rule.is_default else "false" }}</IsDefault>
        <Conditions>
          {% for condition in rule.conditions %}
          <member>
            <Field>{{ condition["field"] }}</Field>
            <Values>
              {% for value in condition["values"] %}
              <member>{{ value }}</member>
              {% endfor %}
            </Values>
          </member>
          {% endfor %}
        </Conditions>
        <Priority>{{ rule.priority }}</Priority>
        <Actions>
          {% for action in rule.actions %}
          <member>
            <Type>{{ action["type"] }}</Type>
            <TargetGroupArn>{{ action["target_group_arn"] }}</TargetGroupArn>
          </member>
          {% endfor %}
        </Actions>
        <RuleArn>{{ rule.arn }}</RuleArn>
      </member>
      {% endfor %}
    </Rules>
  </SetRulePrioritiesResult>
  <ResponseMetadata>
    <RequestId>4d7a8036-f3a7-11e5-9c02-8fd20490d5a6</RequestId>
  </ResponseMetadata>
</SetRulePrioritiesResponse>"""



from __future__ import unicode_literals

import copy

# Ensure 'assert_raises' context manager support for Python 2.6
import tests.backport_assert_raises  # noqa
from nose.tools import assert_raises

import boto3
import boto
from botocore.exceptions import ClientError
from boto.exception import EC2ResponseError
import sure  # noqa

from moto import mock_ec2, mock_ec2_deprecated


@mock_ec2_deprecated
def test_create_and_describe_security_group():
    conn = boto.connect_ec2('the_key', 'the_secret')

    with assert_raises(EC2ResponseError) as ex:
        security_group = conn.create_security_group(
            'test security group', 'this is a test security group', dry_run=True)
    ex.exception.error_code.should.equal('DryRunOperation')
    ex.exception.status.should.equal(400)
    ex.exception.message.should.equal(
        'An error occurred (DryRunOperation) when calling the CreateSecurityGroup operation: Request would have succeeded, but DryRun flag is set')

    security_group = conn.create_security_group(
        'test security group', 'this is a test security group')

    security_group.name.should.equal('test security group')
    security_group.description.should.equal('this is a test security group')

    # Trying to create another group with the same name should throw an error
    with assert_raises(EC2ResponseError) as cm:
        conn.create_security_group(
            'test security group', 'this is a test security group')
    cm.exception.code.should.equal('InvalidGroup.Duplicate')
    cm.exception.status.should.equal(400)
    cm.exception.request_id.should_not.be.none

    all_groups = conn.get_all_security_groups()
    # The default group gets created automatically
    all_groups.should.have.length_of(3)
    group_names = [group.name for group in all_groups]
    set(group_names).should.equal(set(["default", "test security group"]))


@mock_ec2_deprecated
def test_create_security_group_without_description_raises_error():
    conn = boto.connect_ec2('the_key', 'the_secret')

    with assert_raises(EC2ResponseError) as cm:
        conn.create_security_group('test security group', '')
    cm.exception.code.should.equal('MissingParameter')
    cm.exception.status.should.equal(400)
    cm.exception.request_id.should_not.be.none


@mock_ec2_deprecated
def test_default_security_group():
    conn = boto.ec2.connect_to_region('us-east-1')
    groups = conn.get_all_security_groups()
    groups.should.have.length_of(2)
    groups[0].name.should.equal("default")


@mock_ec2_deprecated
def test_create_and_describe_vpc_security_group():
    conn = boto.connect_ec2('the_key', 'the_secret')
    vpc_id = 'vpc-5300000c'
    security_group = conn.create_security_group(
        'test security group', 'this is a test security group', vpc_id=vpc_id)

    security_group.vpc_id.should.equal(vpc_id)

    security_group.name.should.equal('test security group')
    security_group.description.should.equal('this is a test security group')

    # Trying to create another group with the same name in the same VPC should
    # throw an error
    with assert_raises(EC2ResponseError) as cm:
        conn.create_security_group(
            'test security group', 'this is a test security group', vpc_id)
    cm.exception.code.should.equal('InvalidGroup.Duplicate')
    cm.exception.status.should.equal(400)
    cm.exception.request_id.should_not.be.none

    all_groups = conn.get_all_security_groups(filters={'vpc_id': [vpc_id]})

    all_groups[0].vpc_id.should.equal(vpc_id)

    all_groups.should.have.length_of(1)
    all_groups[0].name.should.equal('test security group')


@mock_ec2_deprecated
def test_create_two_security_groups_with_same_name_in_different_vpc():
    conn = boto.connect_ec2('the_key', 'the_secret')
    vpc_id = 'vpc-5300000c'
    vpc_id2 = 'vpc-5300000d'

    conn.create_security_group(
        'test security group', 'this is a test security group', vpc_id)
    conn.create_security_group(
        'test security group', 'this is a test security group', vpc_id2)

    all_groups = conn.get_all_security_groups()

    all_groups.should.have.length_of(4)
    group_names = [group.name for group in all_groups]
    # The default group is created automatically
    set(group_names).should.equal(set(["default", "test security group"]))


@mock_ec2_deprecated
def test_deleting_security_groups():
    conn = boto.connect_ec2('the_key', 'the_secret')
    security_group1 = conn.create_security_group('test1', 'test1')
    conn.create_security_group('test2', 'test2')

    conn.get_all_security_groups().should.have.length_of(4)

    # Deleting a group that doesn't exist should throw an error
    with assert_raises(EC2ResponseError) as cm:
        conn.delete_security_group('foobar')
    cm.exception.code.should.equal('InvalidGroup.NotFound')
    cm.exception.status.should.equal(400)
    cm.exception.request_id.should_not.be.none

    # Delete by name
    with assert_raises(EC2ResponseError) as ex:
        conn.delete_security_group('test2', dry_run=True)
    ex.exception.error_code.should.equal('DryRunOperation')
    ex.exception.status.should.equal(400)
    ex.exception.message.should.equal(
        'An error occurred (DryRunOperation) when calling the DeleteSecurityGroup operation: Request would have succeeded, but DryRun flag is set')

    conn.delete_security_group('test2')
    conn.get_all_security_groups().should.have.length_of(3)

    # Delete by group id
    conn.delete_security_group(group_id=security_group1.id)
    conn.get_all_security_groups().should.have.length_of(2)


@mock_ec2_deprecated
def test_delete_security_group_in_vpc():
    conn = boto.connect_ec2('the_key', 'the_secret')
    vpc_id = "vpc-12345"
    security_group1 = conn.create_security_group('test1', 'test1', vpc_id)

    # this should not throw an exception
    conn.delete_security_group(group_id=security_group1.id)


@mock_ec2_deprecated
def test_authorize_ip_range_and_revoke():
    conn = boto.connect_ec2('the_key', 'the_secret')
    security_group = conn.create_security_group('test', 'test')

    with assert_raises(EC2ResponseError) as ex:
        success = security_group.authorize(
            ip_protocol="tcp", from_port="22", to_port="2222", cidr_ip="123.123.123.123/32", dry_run=True)
    ex.exception.error_code.should.equal('DryRunOperation')
    ex.exception.status.should.equal(400)
    ex.exception.message.should.equal(
        'An error occurred (DryRunOperation) when calling the GrantSecurityGroupIngress operation: Request would have succeeded, but DryRun flag is set')

    success = security_group.authorize(
        ip_protocol="tcp", from_port="22", to_port="2222", cidr_ip="123.123.123.123/32")
    assert success.should.be.true

    security_group = conn.get_all_security_groups(groupnames=['test'])[0]
    int(security_group.rules[0].to_port).should.equal(2222)
    security_group.rules[0].grants[
        0].cidr_ip.should.equal("123.123.123.123/32")

    # Wrong Cidr should throw error
    with assert_raises(EC2ResponseError) as cm:
        security_group.revoke(ip_protocol="tcp", from_port="22",
                              to_port="2222", cidr_ip="123.123.123.122/32")
    cm.exception.code.should.equal('InvalidPermission.NotFound')
    cm.exception.status.should.equal(400)
    cm.exception.request_id.should_not.be.none

    # Actually revoke
    with assert_raises(EC2ResponseError) as ex:
        security_group.revoke(ip_protocol="tcp", from_port="22",
                              to_port="2222", cidr_ip="123.123.123.123/32", dry_run=True)
    ex.exception.error_code.should.equal('DryRunOperation')
    ex.exception.status.should.equal(400)
    ex.exception.message.should.equal(
        'An error occurred (DryRunOperation) when calling the RevokeSecurityGroupIngress operation: Request would have succeeded, but DryRun flag is set')

    security_group.revoke(ip_protocol="tcp", from_port="22",
                          to_port="2222", cidr_ip="123.123.123.123/32")

    security_group = conn.get_all_security_groups()[0]
    security_group.rules.should.have.length_of(0)

    # Test for egress as well
    egress_security_group = conn.create_security_group(
        'testegress', 'testegress', vpc_id='vpc-3432589')

    with assert_raises(EC2ResponseError) as ex:
        success = conn.authorize_security_group_egress(
            egress_security_group.id, "tcp", from_port="22", to_port="2222", cidr_ip="123.123.123.123/32", dry_run=True)
    ex.exception.error_code.should.equal('DryRunOperation')
    ex.exception.status.should.equal(400)
    ex.exception.message.should.equal(
        'An error occurred (DryRunOperation) when calling the GrantSecurityGroupEgress operation: Request would have succeeded, but DryRun flag is set')

    success = conn.authorize_security_group_egress(
        egress_security_group.id, "tcp", from_port="22", to_port="2222", cidr_ip="123.123.123.123/32")
    assert success.should.be.true
    egress_security_group = conn.get_all_security_groups(
        groupnames='testegress')[0]
    # There are two egress rules associated with the security group:
    # the default outbound rule and the new one
    int(egress_security_group.rules_egress[1].to_port).should.equal(2222)
    egress_security_group.rules_egress[1].grants[
        0].cidr_ip.should.equal("123.123.123.123/32")

    # Wrong Cidr should throw error
    egress_security_group.revoke.when.called_with(
        ip_protocol="tcp", from_port="22", to_port="2222", cidr_ip="123.123.123.122/32").should.throw(EC2ResponseError)

    # Actually revoke
    with assert_raises(EC2ResponseError) as ex:
        conn.revoke_security_group_egress(
            egress_security_group.id, "tcp", from_port="22", to_port="2222", cidr_ip="123.123.123.123/32", dry_run=True)
    ex.exception.error_code.should.equal('DryRunOperation')
    ex.exception.status.should.equal(400)
    ex.exception.message.should.equal(
        'An error occurred (DryRunOperation) when calling the RevokeSecurityGroupEgress operation: Request would have succeeded, but DryRun flag is set')

    conn.revoke_security_group_egress(
        egress_security_group.id, "tcp", from_port="22", to_port="2222", cidr_ip="123.123.123.123/32")

    egress_security_group = conn.get_all_security_groups()[0]
    # There is still the default outbound rule
    egress_security_group.rules_egress.should.have.length_of(1)


@mock_ec2_deprecated
def test_authorize_other_group_and_revoke():
    conn = boto.connect_ec2('the_key', 'the_secret')
    security_group = conn.create_security_group('test', 'test')
    other_security_group = conn.create_security_group('other', 'other')
    wrong_group = conn.create_security_group('wrong', 'wrong')

    success = security_group.authorize(
        ip_protocol="tcp", from_port="22", to_port="2222", src_group=other_security_group)
    assert success.should.be.true

    security_group = [
        group for group in conn.get_all_security_groups() if group.name == 'test'][0]
    int(security_group.rules[0].to_port).should.equal(2222)
    security_group.rules[0].grants[
        0].group_id.should.equal(other_security_group.id)

    # Wrong source group should throw error
    with assert_raises(EC2ResponseError) as cm:
        security_group.revoke(ip_protocol="tcp", from_port="22",
                              to_port="2222", src_group=wrong_group)
    cm.exception.code.should.equal('InvalidPermission.NotFound')
    cm.exception.status.should.equal(400)
    cm.exception.request_id.should_not.be.none

    # Actually revoke
    security_group.revoke(ip_protocol="tcp", from_port="22",
                          to_port="2222", src_group=other_security_group)

    security_group = [
        group for group in conn.get_all_security_groups() if group.name == 'test'][0]
    security_group.rules.should.have.length_of(0)


@mock_ec2
def test_authorize_other_group_egress_and_revoke():
    ec2 = boto3.resource('ec2', region_name='us-west-1')

    vpc = ec2.create_vpc(CidrBlock='10.0.0.0/16')

    sg01 = ec2.create_security_group(
        GroupName='sg01', Description='Test security group sg01', VpcId=vpc.id)
    sg02 = ec2.create_security_group(
        GroupName='sg02', Description='Test security group sg02', VpcId=vpc.id)

    ip_permission = {
        'IpProtocol': 'tcp',
        'FromPort': 27017,
        'ToPort': 27017,
        'UserIdGroupPairs': [{'GroupId': sg02.id, 'GroupName': 'sg02', 'UserId': sg02.owner_id}],
        'IpRanges': []
    }

    sg01.authorize_egress(IpPermissions=[ip_permission])
    sg01.ip_permissions_egress.should.have.length_of(2)
    sg01.ip_permissions_egress.should.contain(ip_permission)

    sg01.revoke_egress(IpPermissions=[ip_permission])
    sg01.ip_permissions_egress.should.have.length_of(1)


@mock_ec2_deprecated
def test_authorize_group_in_vpc():
    conn = boto.connect_ec2('the_key', 'the_secret')
    vpc_id = "vpc-12345"

    # create 2 groups in a vpc
    security_group = conn.create_security_group('test1', 'test1', vpc_id)
    other_security_group = conn.create_security_group('test2', 'test2', vpc_id)

    success = security_group.authorize(
        ip_protocol="tcp", from_port="22", to_port="2222", src_group=other_security_group)
    success.should.be.true

    # Check that the rule is accurate
    security_group = [
        group for group in conn.get_all_security_groups() if group.name == 'test1'][0]
    int(security_group.rules[0].to_port).should.equal(2222)
    security_group.rules[0].grants[
        0].group_id.should.equal(other_security_group.id)

    # Now remove the rule
    success = security_group.revoke(
        ip_protocol="tcp", from_port="22", to_port="2222", src_group=other_security_group)
    success.should.be.true

    # And check that it gets revoked
    security_group = [
        group for group in conn.get_all_security_groups() if group.name == 'test1'][0]
    security_group.rules.should.have.length_of(0)


@mock_ec2_deprecated
def test_get_all_security_groups():
    conn = boto.connect_ec2()
    sg1 = conn.create_security_group(
        name='test1', description='test1', vpc_id='vpc-mjm05d27')
    conn.create_security_group(name='test2', description='test2')

    resp = conn.get_all_security_groups(groupnames=['test1'])
    resp.should.have.length_of(1)
    resp[0].id.should.equal(sg1.id)

    with assert_raises(EC2ResponseError) as cm:
        conn.get_all_security_groups(groupnames=['does_not_exist'])
    cm.exception.code.should.equal('InvalidGroup.NotFound')
    cm.exception.status.should.equal(400)
    cm.exception.request_id.should_not.be.none

    resp.should.have.length_of(1)
    resp[0].id.should.equal(sg1.id)

    resp = conn.get_all_security_groups(filters={'vpc-id': ['vpc-mjm05d27']})
    resp.should.have.length_of(1)
    resp[0].id.should.equal(sg1.id)

    resp = conn.get_all_security_groups(filters={'vpc_id': ['vpc-mjm05d27']})
    resp.should.have.length_of(1)
    resp[0].id.should.equal(sg1.id)

    resp = conn.get_all_security_groups(filters={'description': ['test1']})
    resp.should.have.length_of(1)
    resp[0].id.should.equal(sg1.id)

    resp = conn.get_all_security_groups()
    resp.should.have.length_of(4)


@mock_ec2_deprecated
def test_authorize_bad_cidr_throws_invalid_parameter_value():
    conn = boto.connect_ec2('the_key', 'the_secret')
    security_group = conn.create_security_group('test', 'test')
    with assert_raises(EC2ResponseError) as cm:
        security_group.authorize(
            ip_protocol="tcp", from_port="22", to_port="2222", cidr_ip="123.123.123.123")
    cm.exception.code.should.equal('InvalidParameterValue')
    cm.exception.status.should.equal(400)
    cm.exception.request_id.should_not.be.none


@mock_ec2_deprecated
def test_security_group_tagging():
    conn = boto.connect_vpc()
    vpc = conn.create_vpc("10.0.0.0/16")

    sg = conn.create_security_group("test-sg", "Test SG", vpc.id)

    with assert_raises(EC2ResponseError) as ex:
        sg.add_tag("Test", "Tag", dry_run=True)
    ex.exception.error_code.should.equal('DryRunOperation')
    ex.exception.status.should.equal(400)
    ex.exception.message.should.equal(
        'An error occurred (DryRunOperation) when calling the CreateTags operation: Request would have succeeded, but DryRun flag is set')

    sg.add_tag("Test", "Tag")

    tag = conn.get_all_tags()[0]
    tag.name.should.equal("Test")
    tag.value.should.equal("Tag")

    group = conn.get_all_security_groups("test-sg")[0]
    group.tags.should.have.length_of(1)
    group.tags["Test"].should.equal("Tag")


@mock_ec2_deprecated
def test_security_group_tag_filtering():
    conn = boto.connect_ec2()
    sg = conn.create_security_group("test-sg", "Test SG")
    sg.add_tag("test-tag", "test-value")

    groups = conn.get_all_security_groups(
        filters={"tag:test-tag": "test-value"})
    groups.should.have.length_of(1)


@mock_ec2_deprecated
def test_authorize_all_protocols_with_no_port_specification():
    conn = boto.connect_ec2()
    sg = conn.create_security_group('test', 'test')

    success = sg.authorize(ip_protocol='-1', cidr_ip='0.0.0.0/0')
    success.should.be.true

    sg = conn.get_all_security_groups('test')[0]
    sg.rules[0].from_port.should.equal(None)
    sg.rules[0].to_port.should.equal(None)


@mock_ec2_deprecated
def test_sec_group_rule_limit():
    ec2_conn = boto.connect_ec2()
    sg = ec2_conn.create_security_group('test', 'test')
    other_sg = ec2_conn.create_security_group('test_2', 'test_other')

    # INGRESS
    with assert_raises(EC2ResponseError) as cm:
        ec2_conn.authorize_security_group(
            group_id=sg.id, ip_protocol='-1',
            cidr_ip=['{0}.0.0.0/0'.format(i) for i in range(110)])
    cm.exception.error_code.should.equal('RulesPerSecurityGroupLimitExceeded')

    sg.rules.should.be.empty
    # authorize a rule targeting a different sec group (because this count too)
    success = ec2_conn.authorize_security_group(
        group_id=sg.id, ip_protocol='-1',
        src_security_group_group_id=other_sg.id)
    success.should.be.true
    # fill the rules up the limit
    success = ec2_conn.authorize_security_group(
        group_id=sg.id, ip_protocol='-1',
        cidr_ip=['{0}.0.0.0/0'.format(i) for i in range(99)])
    success.should.be.true
    # verify that we cannot authorize past the limit for a CIDR IP
    with assert_raises(EC2ResponseError) as cm:
        ec2_conn.authorize_security_group(
            group_id=sg.id, ip_protocol='-1', cidr_ip=['100.0.0.0/0'])
    cm.exception.error_code.should.equal('RulesPerSecurityGroupLimitExceeded')
    # verify that we cannot authorize past the limit for a different sec group
    with assert_raises(EC2ResponseError) as cm:
        ec2_conn.authorize_security_group(
            group_id=sg.id, ip_protocol='-1',
            src_security_group_group_id=other_sg.id)
    cm.exception.error_code.should.equal('RulesPerSecurityGroupLimitExceeded')

    # EGRESS
    # authorize a rule targeting a different sec group (because this count too)
    ec2_conn.authorize_security_group_egress(
        group_id=sg.id, ip_protocol='-1',
        src_group_id=other_sg.id)
    # fill the rules up the limit
    # remember that by default, when created a sec group contains 1 egress rule
    # so our other_sg rule + 98 CIDR IP rules + 1 by default == 100 the limit
    for i in range(98):
        ec2_conn.authorize_security_group_egress(
            group_id=sg.id, ip_protocol='-1',
            cidr_ip='{0}.0.0.0/0'.format(i))
    # verify that we cannot authorize past the limit for a CIDR IP
    with assert_raises(EC2ResponseError) as cm:
        ec2_conn.authorize_security_group_egress(
            group_id=sg.id, ip_protocol='-1',
            cidr_ip='101.0.0.0/0')
    cm.exception.error_code.should.equal('RulesPerSecurityGroupLimitExceeded')
    # verify that we cannot authorize past the limit for a different sec group
    with assert_raises(EC2ResponseError) as cm:
        ec2_conn.authorize_security_group_egress(
            group_id=sg.id, ip_protocol='-1',
            src_group_id=other_sg.id)
    cm.exception.error_code.should.equal('RulesPerSecurityGroupLimitExceeded')


@mock_ec2_deprecated
def test_sec_group_rule_limit_vpc():
    ec2_conn = boto.connect_ec2()
    vpc_conn = boto.connect_vpc()

    vpc = vpc_conn.create_vpc('10.0.0.0/8')

    sg = ec2_conn.create_security_group('test', 'test', vpc_id=vpc.id)
    other_sg = ec2_conn.create_security_group('test_2', 'test', vpc_id=vpc.id)

    # INGRESS
    with assert_raises(EC2ResponseError) as cm:
        ec2_conn.authorize_security_group(
            group_id=sg.id, ip_protocol='-1',
            cidr_ip=['{0}.0.0.0/0'.format(i) for i in range(110)])
    cm.exception.error_code.should.equal('RulesPerSecurityGroupLimitExceeded')

    sg.rules.should.be.empty
    # authorize a rule targeting a different sec group (because this count too)
    success = ec2_conn.authorize_security_group(
        group_id=sg.id, ip_protocol='-1',
        src_security_group_group_id=other_sg.id)
    success.should.be.true
    # fill the rules up the limit
    success = ec2_conn.authorize_security_group(
        group_id=sg.id, ip_protocol='-1',
        cidr_ip=['{0}.0.0.0/0'.format(i) for i in range(49)])
    # verify that we cannot authorize past the limit for a CIDR IP
    success.should.be.true
    with assert_raises(EC2ResponseError) as cm:
        ec2_conn.authorize_security_group(
            group_id=sg.id, ip_protocol='-1', cidr_ip=['100.0.0.0/0'])
    cm.exception.error_code.should.equal('RulesPerSecurityGroupLimitExceeded')
    # verify that we cannot authorize past the limit for a different sec group
    with assert_raises(EC2ResponseError) as cm:
        ec2_conn.authorize_security_group(
            group_id=sg.id, ip_protocol='-1',
            src_security_group_group_id=other_sg.id)
    cm.exception.error_code.should.equal('RulesPerSecurityGroupLimitExceeded')

    # EGRESS
    # authorize a rule targeting a different sec group (because this count too)
    ec2_conn.authorize_security_group_egress(
        group_id=sg.id, ip_protocol='-1',
        src_group_id=other_sg.id)
    # fill the rules up the limit
    # remember that by default, when created a sec group contains 1 egress rule
    # so our other_sg rule + 48 CIDR IP rules + 1 by default == 50 the limit
    for i in range(48):
        ec2_conn.authorize_security_group_egress(
            group_id=sg.id, ip_protocol='-1',
            cidr_ip='{0}.0.0.0/0'.format(i))
    # verify that we cannot authorize past the limit for a CIDR IP
    with assert_raises(EC2ResponseError) as cm:
        ec2_conn.authorize_security_group_egress(
            group_id=sg.id, ip_protocol='-1',
            cidr_ip='50.0.0.0/0')
    cm.exception.error_code.should.equal('RulesPerSecurityGroupLimitExceeded')
    # verify that we cannot authorize past the limit for a different sec group
    with assert_raises(EC2ResponseError) as cm:
        ec2_conn.authorize_security_group_egress(
            group_id=sg.id, ip_protocol='-1',
            src_group_id=other_sg.id)
    cm.exception.error_code.should.equal('RulesPerSecurityGroupLimitExceeded')


'''
Boto3
'''


@mock_ec2
def test_add_same_rule_twice_throws_error():
    ec2 = boto3.resource('ec2', region_name='us-west-1')

    vpc = ec2.create_vpc(CidrBlock='10.0.0.0/16')
    sg = ec2.create_security_group(
        GroupName='sg1', Description='Test security group sg1', VpcId=vpc.id)

    ip_permissions = [
        {
            'IpProtocol': 'tcp',
            'FromPort': 27017,
            'ToPort': 27017,
            'IpRanges': [{"CidrIp": "1.2.3.4/32"}]
        },
    ]
    sg.authorize_ingress(IpPermissions=ip_permissions)

    with assert_raises(ClientError) as ex:
        sg.authorize_ingress(IpPermissions=ip_permissions)


@mock_ec2
def test_security_group_tagging_boto3():
    conn = boto3.client('ec2', region_name='us-east-1')

    sg = conn.create_security_group(GroupName="test-sg", Description="Test SG")

    with assert_raises(ClientError) as ex:
        conn.create_tags(Resources=[sg['GroupId']], Tags=[
                         {'Key': 'Test', 'Value': 'Tag'}], DryRun=True)
    ex.exception.response['Error']['Code'].should.equal('DryRunOperation')
    ex.exception.response['ResponseMetadata'][
        'HTTPStatusCode'].should.equal(400)
    ex.exception.response['Error']['Message'].should.equal(
        'An error occurred (DryRunOperation) when calling the CreateTags operation: Request would have succeeded, but DryRun flag is set')

    conn.create_tags(Resources=[sg['GroupId']], Tags=[
                     {'Key': 'Test', 'Value': 'Tag'}])
    describe = conn.describe_security_groups(
        Filters=[{'Name': 'tag-value', 'Values': ['Tag']}])
    tag = describe["SecurityGroups"][0]['Tags'][0]
    tag['Value'].should.equal("Tag")
    tag['Key'].should.equal("Test")


@mock_ec2
def test_authorize_and_revoke_in_bulk():
    ec2 = boto3.resource('ec2', region_name='us-west-1')

    vpc = ec2.create_vpc(CidrBlock='10.0.0.0/16')

    sg01 = ec2.create_security_group(
        GroupName='sg01', Description='Test security group sg01', VpcId=vpc.id)
    sg02 = ec2.create_security_group(
        GroupName='sg02', Description='Test security group sg02', VpcId=vpc.id)
    sg03 = ec2.create_security_group(
        GroupName='sg03', Description='Test security group sg03')

    ip_permissions = [
        {
            'IpProtocol': 'tcp',
            'FromPort': 27017,
            'ToPort': 27017,
            'UserIdGroupPairs': [{'GroupId': sg02.id, 'GroupName': 'sg02',
                                  'UserId': sg02.owner_id}],
            'IpRanges': []
        },
        {
            'IpProtocol': 'tcp',
            'FromPort': 27018,
            'ToPort': 27018,
            'UserIdGroupPairs': [{'GroupId': sg02.id, 'UserId': sg02.owner_id}],
            'IpRanges': []
        },
        {
            'IpProtocol': 'tcp',
            'FromPort': 27017,
            'ToPort': 27017,
            'UserIdGroupPairs': [{'GroupName': 'sg03', 'UserId': sg03.owner_id}],
            'IpRanges': []
        }
    ]
    expected_ip_permissions = copy.deepcopy(ip_permissions)
    expected_ip_permissions[1]['UserIdGroupPairs'][0]['GroupName'] = 'sg02'
    expected_ip_permissions[2]['UserIdGroupPairs'][0]['GroupId'] = sg03.id

    sg01.authorize_ingress(IpPermissions=ip_permissions)
    sg01.ip_permissions.should.have.length_of(3)
    for ip_permission in expected_ip_permissions:
        sg01.ip_permissions.should.contain(ip_permission)

    sg01.revoke_ingress(IpPermissions=ip_permissions)
    sg01.ip_permissions.should.be.empty
    for ip_permission in expected_ip_permissions:
        sg01.ip_permissions.shouldnt.contain(ip_permission)

    sg01.authorize_egress(IpPermissions=ip_permissions)
    sg01.ip_permissions_egress.should.have.length_of(4)
    for ip_permission in expected_ip_permissions:
        sg01.ip_permissions_egress.should.contain(ip_permission)

    sg01.revoke_egress(IpPermissions=ip_permissions)
    sg01.ip_permissions_egress.should.have.length_of(1)
    for ip_permission in expected_ip_permissions:
        sg01.ip_permissions_egress.shouldnt.contain(ip_permission)


@mock_ec2_deprecated
def test_get_all_security_groups_filter_with_same_vpc_id():
    conn = boto.connect_ec2('the_key', 'the_secret')
    vpc_id = 'vpc-5300000c'
    security_group = conn.create_security_group(
        'test1', 'test1', vpc_id=vpc_id)
    security_group2 = conn.create_security_group(
        'test2', 'test2', vpc_id=vpc_id)

    security_group.vpc_id.should.equal(vpc_id)
    security_group2.vpc_id.should.equal(vpc_id)

    security_groups = conn.get_all_security_groups(
        group_ids=[security_group.id], filters={'vpc-id': [vpc_id]})
    security_groups.should.have.length_of(1)

    with assert_raises(EC2ResponseError) as cm:
        conn.get_all_security_groups(group_ids=['does_not_exist'])
    cm.exception.code.should.equal('InvalidGroup.NotFound')
    cm.exception.status.should.equal(400)
    cm.exception.request_id.should_not.be.none



from __future__ import unicode_literals

import boto
from boto.exception import BotoServerError
from moto import mock_sns_deprecated
import sure  # noqa


@mock_sns_deprecated
def test_create_platform_application():
    conn = boto.connect_sns()
    platform_application = conn.create_platform_application(
        name="my-application",
        platform="APNS",
        attributes={
            "PlatformCredential": "platform_credential",
            "PlatformPrincipal": "platform_principal",
        },
    )
    application_arn = platform_application['CreatePlatformApplicationResponse'][
        'CreatePlatformApplicationResult']['PlatformApplicationArn']
    application_arn.should.equal(
        'arn:aws:sns:us-east-1:123456789012:app/APNS/my-application')


@mock_sns_deprecated
def test_get_platform_application_attributes():
    conn = boto.connect_sns()
    platform_application = conn.create_platform_application(
        name="my-application",
        platform="APNS",
        attributes={
            "PlatformCredential": "platform_credential",
            "PlatformPrincipal": "platform_principal",
        },
    )
    arn = platform_application['CreatePlatformApplicationResponse'][
        'CreatePlatformApplicationResult']['PlatformApplicationArn']
    attributes = conn.get_platform_application_attributes(arn)['GetPlatformApplicationAttributesResponse'][
        'GetPlatformApplicationAttributesResult']['Attributes']
    attributes.should.equal({
        "PlatformCredential": "platform_credential",
        "PlatformPrincipal": "platform_principal",
    })


@mock_sns_deprecated
def test_get_missing_platform_application_attributes():
    conn = boto.connect_sns()
    conn.get_platform_application_attributes.when.called_with(
        "a-fake-arn").should.throw(BotoServerError)


@mock_sns_deprecated
def test_set_platform_application_attributes():
    conn = boto.connect_sns()
    platform_application = conn.create_platform_application(
        name="my-application",
        platform="APNS",
        attributes={
            "PlatformCredential": "platform_credential",
            "PlatformPrincipal": "platform_principal",
        },
    )
    arn = platform_application['CreatePlatformApplicationResponse'][
        'CreatePlatformApplicationResult']['PlatformApplicationArn']
    conn.set_platform_application_attributes(arn,
                                             {"PlatformPrincipal": "other"}
                                             )
    attributes = conn.get_platform_application_attributes(arn)['GetPlatformApplicationAttributesResponse'][
        'GetPlatformApplicationAttributesResult']['Attributes']
    attributes.should.equal({
        "PlatformCredential": "platform_credential",
        "PlatformPrincipal": "other",
    })


@mock_sns_deprecated
def test_list_platform_applications():
    conn = boto.connect_sns()
    conn.create_platform_application(
        name="application1",
        platform="APNS",
    )
    conn.create_platform_application(
        name="application2",
        platform="APNS",
    )

    applications_repsonse = conn.list_platform_applications()
    applications = applications_repsonse['ListPlatformApplicationsResponse'][
        'ListPlatformApplicationsResult']['PlatformApplications']
    applications.should.have.length_of(2)


@mock_sns_deprecated
def test_delete_platform_application():
    conn = boto.connect_sns()
    conn.create_platform_application(
        name="application1",
        platform="APNS",
    )
    conn.create_platform_application(
        name="application2",
        platform="APNS",
    )

    applications_repsonse = conn.list_platform_applications()
    applications = applications_repsonse['ListPlatformApplicationsResponse'][
        'ListPlatformApplicationsResult']['PlatformApplications']
    applications.should.have.length_of(2)

    application_arn = applications[0]['PlatformApplicationArn']
    conn.delete_platform_application(application_arn)

    applications_repsonse = conn.list_platform_applications()
    applications = applications_repsonse['ListPlatformApplicationsResponse'][
        'ListPlatformApplicationsResult']['PlatformApplications']
    applications.should.have.length_of(1)


@mock_sns_deprecated
def test_create_platform_endpoint():
    conn = boto.connect_sns()
    platform_application = conn.create_platform_application(
        name="my-application",
        platform="APNS",
    )
    application_arn = platform_application['CreatePlatformApplicationResponse'][
        'CreatePlatformApplicationResult']['PlatformApplicationArn']

    endpoint = conn.create_platform_endpoint(
        platform_application_arn=application_arn,
        token="some_unique_id",
        custom_user_data="some user data",
        attributes={
            "Enabled": False,
        },
    )

    endpoint_arn = endpoint['CreatePlatformEndpointResponse'][
        'CreatePlatformEndpointResult']['EndpointArn']
    endpoint_arn.should.contain(
        "arn:aws:sns:us-east-1:123456789012:endpoint/APNS/my-application/")


@mock_sns_deprecated
def test_get_list_endpoints_by_platform_application():
    conn = boto.connect_sns()
    platform_application = conn.create_platform_application(
        name="my-application",
        platform="APNS",
    )
    application_arn = platform_application['CreatePlatformApplicationResponse'][
        'CreatePlatformApplicationResult']['PlatformApplicationArn']

    endpoint = conn.create_platform_endpoint(
        platform_application_arn=application_arn,
        token="some_unique_id",
        custom_user_data="some user data",
        attributes={
            "CustomUserData": "some data",
        },
    )
    endpoint_arn = endpoint['CreatePlatformEndpointResponse'][
        'CreatePlatformEndpointResult']['EndpointArn']

    endpoint_list = conn.list_endpoints_by_platform_application(
        platform_application_arn=application_arn
    )['ListEndpointsByPlatformApplicationResponse']['ListEndpointsByPlatformApplicationResult']['Endpoints']

    endpoint_list.should.have.length_of(1)
    endpoint_list[0]['Attributes']['CustomUserData'].should.equal('some data')
    endpoint_list[0]['EndpointArn'].should.equal(endpoint_arn)


@mock_sns_deprecated
def test_get_endpoint_attributes():
    conn = boto.connect_sns()
    platform_application = conn.create_platform_application(
        name="my-application",
        platform="APNS",
    )
    application_arn = platform_application['CreatePlatformApplicationResponse'][
        'CreatePlatformApplicationResult']['PlatformApplicationArn']

    endpoint = conn.create_platform_endpoint(
        platform_application_arn=application_arn,
        token="some_unique_id",
        custom_user_data="some user data",
        attributes={
            "Enabled": False,
            "CustomUserData": "some data",
        },
    )
    endpoint_arn = endpoint['CreatePlatformEndpointResponse'][
        'CreatePlatformEndpointResult']['EndpointArn']

    attributes = conn.get_endpoint_attributes(endpoint_arn)['GetEndpointAttributesResponse'][
        'GetEndpointAttributesResult']['Attributes']
    attributes.should.equal({
        "Token": "some_unique_id",
        "Enabled": 'False',
        "CustomUserData": "some data",
    })


@mock_sns_deprecated
def test_get_missing_endpoint_attributes():
    conn = boto.connect_sns()
    conn.get_endpoint_attributes.when.called_with(
        "a-fake-arn").should.throw(BotoServerError)


@mock_sns_deprecated
def test_set_endpoint_attributes():
    conn = boto.connect_sns()
    platform_application = conn.create_platform_application(
        name="my-application",
        platform="APNS",
    )
    application_arn = platform_application['CreatePlatformApplicationResponse'][
        'CreatePlatformApplicationResult']['PlatformApplicationArn']

    endpoint = conn.create_platform_endpoint(
        platform_application_arn=application_arn,
        token="some_unique_id",
        custom_user_data="some user data",
        attributes={
            "Enabled": False,
            "CustomUserData": "some data",
        },
    )
    endpoint_arn = endpoint['CreatePlatformEndpointResponse'][
        'CreatePlatformEndpointResult']['EndpointArn']

    conn.set_endpoint_attributes(endpoint_arn,
                                 {"CustomUserData": "other data"}
                                 )
    attributes = conn.get_endpoint_attributes(endpoint_arn)['GetEndpointAttributesResponse'][
        'GetEndpointAttributesResult']['Attributes']
    attributes.should.equal({
        "Token": "some_unique_id",
        "Enabled": 'False',
        "CustomUserData": "other data",
    })


@mock_sns_deprecated
def test_delete_endpoint():
    conn = boto.connect_sns()
    platform_application = conn.create_platform_application(
        name="my-application",
        platform="APNS",
    )
    application_arn = platform_application['CreatePlatformApplicationResponse'][
        'CreatePlatformApplicationResult']['PlatformApplicationArn']

    endpoint = conn.create_platform_endpoint(
        platform_application_arn=application_arn,
        token="some_unique_id",
        custom_user_data="some user data",
        attributes={
            "Enabled": False,
            "CustomUserData": "some data",
        },
    )
    endpoint_arn = endpoint['CreatePlatformEndpointResponse'][
        'CreatePlatformEndpointResult']['EndpointArn']

    endpoint_list = conn.list_endpoints_by_platform_application(
        platform_application_arn=application_arn
    )['ListEndpointsByPlatformApplicationResponse']['ListEndpointsByPlatformApplicationResult']['Endpoints']

    endpoint_list.should.have.length_of(1)

    conn.delete_endpoint(endpoint_arn)

    endpoint_list = conn.list_endpoints_by_platform_application(
        platform_application_arn=application_arn
    )['ListEndpointsByPlatformApplicationResponse']['ListEndpointsByPlatformApplicationResult']['Endpoints']
    endpoint_list.should.have.length_of(0)


@mock_sns_deprecated
def test_publish_to_platform_endpoint():
    conn = boto.connect_sns()
    platform_application = conn.create_platform_application(
        name="my-application",
        platform="APNS",
    )
    application_arn = platform_application['CreatePlatformApplicationResponse'][
        'CreatePlatformApplicationResult']['PlatformApplicationArn']

    endpoint = conn.create_platform_endpoint(
        platform_application_arn=application_arn,
        token="some_unique_id",
        custom_user_data="some user data",
        attributes={
            "Enabled": True,
        },
    )

    endpoint_arn = endpoint['CreatePlatformEndpointResponse'][
        'CreatePlatformEndpointResult']['EndpointArn']

    conn.publish(message="some message", message_structure="json",
                 target_arn=endpoint_arn)



from __future__ import unicode_literals
from mock import patch
import sure  # noqa

from moto.server import main, create_backend_app, DomainDispatcherApplication


def test_wrong_arguments():
    try:
        main(["name", "test1", "test2", "test3"])
        assert False, ("main() when called with the incorrect number of args"
                       " should raise a system exit")
    except SystemExit:
        pass


@patch('moto.server.run_simple')
def test_right_arguments(run_simple):
    main(["s3"])
    func_call = run_simple.call_args[0]
    func_call[0].should.equal("127.0.0.1")
    func_call[1].should.equal(5000)


@patch('moto.server.run_simple')
def test_port_argument(run_simple):
    main(["s3", "--port", "8080"])
    func_call = run_simple.call_args[0]
    func_call[0].should.equal("127.0.0.1")
    func_call[1].should.equal(8080)


def test_domain_dispatched():
    dispatcher = DomainDispatcherApplication(create_backend_app)
    backend_app = dispatcher.get_application(
        {"HTTP_HOST": "email.us-east1.amazonaws.com"})
    keys = list(backend_app.view_functions.keys())
    keys[0].should.equal('EmailResponse.dispatch')


def test_domain_without_matches():
    dispatcher = DomainDispatcherApplication(create_backend_app)
    dispatcher.get_application.when.called_with(
        {"HTTP_HOST": "not-matching-anything.com"}).should.throw(RuntimeError)


def test_domain_dispatched_with_service():
    # If we pass a particular service, always return that.
    dispatcher = DomainDispatcherApplication(create_backend_app, service="s3")
    backend_app = dispatcher.get_application(
        {"HTTP_HOST": "s3.us-east1.amazonaws.com"})
    keys = set(backend_app.view_functions.keys())
    keys.should.contain('ResponseObject.key_response')



from __future__ import unicode_literals

import json
from werkzeug.exceptions import BadRequest


class ResourceNotFoundException(BadRequest):

    def __init__(self, message):
        super(ResourceNotFoundException, self).__init__()
        self.description = json.dumps({
            "message": message,
            '__type': 'ResourceNotFoundException',
        })


class ValidationException(BadRequest):

    def __init__(self, message):
        super(ValidationException, self).__init__()
        self.description = json.dumps({
            "message": message,
            '__type': 'ResourceNotFoundException',
        })



from __future__ import unicode_literals
from moto.core.responses import BaseResponse


class SpotFleets(BaseResponse):

    def cancel_spot_fleet_requests(self):
        spot_fleet_request_ids = self._get_multi_param("SpotFleetRequestId.")
        terminate_instances = self._get_param("TerminateInstances")
        spot_fleets = self.ec2_backend.cancel_spot_fleet_requests(
            spot_fleet_request_ids, terminate_instances)
        template = self.response_template(CANCEL_SPOT_FLEETS_TEMPLATE)
        return template.render(spot_fleets=spot_fleets)

    def describe_spot_fleet_instances(self):
        spot_fleet_request_id = self._get_param("SpotFleetRequestId")

        spot_requests = self.ec2_backend.describe_spot_fleet_instances(
            spot_fleet_request_id)
        template = self.response_template(
            DESCRIBE_SPOT_FLEET_INSTANCES_TEMPLATE)
        return template.render(spot_request_id=spot_fleet_request_id, spot_requests=spot_requests)

    def describe_spot_fleet_requests(self):
        spot_fleet_request_ids = self._get_multi_param("SpotFleetRequestId.")

        requests = self.ec2_backend.describe_spot_fleet_requests(
            spot_fleet_request_ids)
        template = self.response_template(DESCRIBE_SPOT_FLEET_TEMPLATE)
        return template.render(requests=requests)

    def request_spot_fleet(self):
        spot_config = self._get_dict_param("SpotFleetRequestConfig.")
        spot_price = spot_config['spot_price']
        target_capacity = spot_config['target_capacity']
        iam_fleet_role = spot_config['iam_fleet_role']
        allocation_strategy = spot_config['allocation_strategy']

        launch_specs = self._get_list_prefix(
            "SpotFleetRequestConfig.LaunchSpecifications")

        request = self.ec2_backend.request_spot_fleet(
            spot_price=spot_price,
            target_capacity=target_capacity,
            iam_fleet_role=iam_fleet_role,
            allocation_strategy=allocation_strategy,
            launch_specs=launch_specs,
        )

        template = self.response_template(REQUEST_SPOT_FLEET_TEMPLATE)
        return template.render(request=request)


REQUEST_SPOT_FLEET_TEMPLATE = """<RequestSpotFleetResponse xmlns="http://ec2.amazonaws.com/doc/2016-09-15/">
    <requestId>60262cc5-2bd4-4c8d-98ed-example</requestId>
    <spotFleetRequestId>{{ request.id }}</spotFleetRequestId>
</RequestSpotFleetResponse>"""

DESCRIBE_SPOT_FLEET_TEMPLATE = """<DescribeSpotFleetRequestsResponse xmlns="http://ec2.amazonaws.com/doc/2016-09-15/">
    <requestId>4d68a6cc-8f2e-4be1-b425-example</requestId>
    <spotFleetRequestConfigSet>
        {% for request in requests %}
        <item>
            <spotFleetRequestId>{{ request.id }}</spotFleetRequestId>
            <spotFleetRequestState>{{ request.state }}</spotFleetRequestState>
            <spotFleetRequestConfig>
                <spotPrice>{{ request.spot_price }}</spotPrice>
                <targetCapacity>{{ request.target_capacity }}</targetCapacity>
                <iamFleetRole>{{ request.iam_fleet_role }}</iamFleetRole>
                <allocationStrategy>{{ request.allocation_strategy }}</allocationStrategy>
                <fulfilledCapacity>{{ request.fulfilled_capacity }}</fulfilledCapacity>
                <launchSpecifications>
                    {% for launch_spec in request.launch_specs %}
                    <item>
                        <subnetId>{{ launch_spec.subnet_id }}</subnetId>
                        <ebsOptimized>{{ launch_spec.ebs_optimized }}</ebsOptimized>
                        <imageId>{{ launch_spec.image_id }}</imageId>
                        <instanceType>{{ launch_spec.instance_type }}</instanceType>
                        <iamInstanceProfile><arn>{{ launch_spec.iam_instance_profile }}</arn></iamInstanceProfile>
                        <keyName>{{ launch_spec.key_name }}</keyName>
                        <monitoring><enabled>{{ launch_spec.monitoring }}</enabled></monitoring>
                        <spotPrice>{{ launch_spec.spot_price }}</spotPrice>
                        <userData>{{ launch_spec.user_data }}</userData>
                        <weightedCapacity>{{ launch_spec.weighted_capacity }}</weightedCapacity>
                        <groupSet>
                            {% for group in launch_spec.group_set %}
                            <item>
                                <groupId>{{ group }}</groupId>
                            </item>
                            {% endfor %}
                        </groupSet>
                    </item>
                    {% endfor %}
                </launchSpecifications>
            </spotFleetRequestConfig>
        </item>
        {% endfor %}
    </spotFleetRequestConfigSet>
</DescribeSpotFleetRequestsResponse>"""

DESCRIBE_SPOT_FLEET_INSTANCES_TEMPLATE = """<DescribeSpotFleetInstancesResponse xmlns="http://ec2.amazonaws.com/doc/2016-09-15/">
    <requestId>cfb09950-45e2-472d-a6a9-example</requestId>
    <spotFleetRequestId>{{ spot_request_id }}</spotFleetRequestId>
    <activeInstanceSet>
        {% for spot_request in spot_requests %}
        <item>
            <instanceId>{{ spot_request.instance.id }}</instanceId>
            <spotInstanceRequestId>{{ spot_request.id }}</spotInstanceRequestId>
            <instanceType>{{ spot_request.instance.instance_type }}</instanceType>
        </item>
        {% endfor %}
    </activeInstanceSet>
</DescribeSpotFleetInstancesResponse>
"""

CANCEL_SPOT_FLEETS_TEMPLATE = """<CancelSpotFleetRequestsResponse xmlns="http://ec2.amazonaws.com/doc/2016-09-15/">
    <requestId>e12d2fe5-6503-4b4b-911c-example</requestId>
    <unsuccessfulFleetRequestSet/>
    <successfulFleetRequestSet>
        {% for spot_fleet in spot_fleets %}
        <item>
            <spotFleetRequestId>{{ spot_fleet.id }}</spotFleetRequestId>
            <currentSpotFleetRequestState>cancelled_terminating</currentSpotFleetRequestState>
            <previousSpotFleetRequestState>active</previousSpotFleetRequestState>
        </item>
        {% endfor %}
    </successfulFleetRequestSet>
</CancelSpotFleetRequestsResponse>"""
